r"""
kmltools module: $CLAW/geoclaw/src/python/geoclaw/kmltools.py

Tools to make kml files to overlay on Google Earth.
Note that color is in KML format, BGR with 2 hex digits for each, e.g.

  FF0000 is blue, 00FF00 is green,  0000FF is red, 00FF00 is yellow.

Actually it's an 8 hex digit number, where the first two digits are
transparency, but in this module these default to 'FF' (but you can specify
the full 8 digits if you want it transparent).

:Functions:
 - deg2dms - convert decimal degrees to (degrees, minutes, seconds)
 - regions2kml - create a kml outline for each regions specified in setrun
 - box2kml - create a kml outline from a rectangular box
 - quad2kml - create a kml outline for an arbitrary quadrilateral
 - poly2kml - create a kml outline for an arbitrary polygon
 - line2kml - create a kml line connecting 2 points
 - gauges2kml - create a kml marker for each gauge specified in setrun
 - topo2kml - create a kml outline for each topo grid specified in setrun
 - dtopo2kml - create a kml outline for each dtopo grid specified in setrun
 - fgmax2kml - create a kml outline for each fgmax grid specified in setrun
 - fgout2kml - create a kml outline for each fgout grid specified in setrun
 - make_input_data_kmls - make kml files for many things specified in setrun
 - pcolorcells_for_kml - version of pcolormesh with appropriate dpi and size
 - png2kml - create kml file wrapping a png figure to be viewed on GE
 - kml_build_colorbar - create a colorbar to display on GE
 - topo2kmz - create kmz file showing onshore and offshore topography
 - transect2kml - create kml file showing a set of points on a transect
 - kml_header - used internally
 - kml_footer - used internally
 - kml_region - used internally
 - kml_gauge - used internally
 - kml_png - used internally
 - kml_cb - used internally
"""

try:
    from importlib import reload
except:
    pass # assume python2, already has reload

def f2s(x, num_digits=6):
    r"""
    Convert float to string in fixed point notation with at most
    *num_digits* digits of precision and trailing zeros removed, 
    for printing nicely in kml description boxes.
    """
    format = '%' + '.%sf' % num_digits
    s = (format % x).rstrip('0')
    return s
    
def deg2dms(dy):
    r"""
    Convert decimal degrees to tuple (degrees, minutes, seconds)
    """

    from numpy import floor
    dy_deg = floor(dy)
    dy_min = floor((dy-dy_deg)*60.)
    dy_sec = (dy-dy_deg-dy_min/60.)*3600.
    return dy_deg,dy_min,dy_sec


def regions2kml(rundata=None,fname='regions.kml',verbose=True,combined=True):

    """
    Create a KML box for each AMR region specified for a GeoClaw run.

    :Inputs:

      - *rundata* - an object of class *ClawRunData* or None

        If *rundata==None*, try to create based on executing function *setrun*
        from the `setrun.py` file in the current directory.

      - *fname* (str) - resulting kml file.

      - *verbose* (bool) - If *True*, print out info about each region found

      - *combined* (bool) - If *True*, combine into single kml file with
        name given by *fname*.  This is the default. 
        If False, *fname* is ignored and individual files are created for
        each region with names are Domain.kml, Region00.kml, etc.
        These will show up separately in GoogleEarth so they can be turned
        on or off individually.

    First create a box for the entire domain (in red) and then a box
    for each region (in white).

    :Example:

        >>> from clawpack.geoclaw import kmltools
        >>> kmltools.regions2kml()

    is equivalent to:

        >>> from clawpack.geoclaw import kmltools
        >>> from setrun import setrun
        >>> rundata = setrun()
        >>> kmltools.regions2kml(rundata)

    By default this creates a file named *regions.kml* that can be opened in
    Google Earth.

    """

    from numpy import cos,pi,floor

    if rundata is None:
        try:
            import setrun
            reload(setrun)
            rundata = setrun.setrun()
        except:
            raise IOError("*** cannot execute setrun file")

    fname_combined = 'FlagRegions.kml'
    
    clawdata = rundata.clawdata
    x1,y1 = clawdata.lower[0:]
    x2,y2 = clawdata.upper[0:]
    description = "  x1 = %s, x2 = %s\n" % (f2s(x1),f2s(x2)) \
            + "  y1 = %s, y2 = %s\n" % (f2s(y1),f2s(y2))

    mx,my = clawdata.num_cells[0:]
    dx = (x2-x1)/float(mx)
    dx_meters = dx*111e3*cos(pi*0.5*(y1+y2)/180.)
    dy = (y2-y1)/float(my)
    dy_meters = dy*111e3
    if verbose:
        print("Domain:   %10.6f  %10.6f  %10.6f  %10.6f" % (x1,x2,y1,y2))
    dx_deg,dx_min,dx_sec = deg2dms(dx)
    dy_deg,dy_min,dy_sec = deg2dms(dy)
    #print "Level 1 resolution:  dx = %g deg, %g min, %g sec = %g meters" \
    #   % (dx_deg,dx_min,dx_sec,dx_meters)
    levtext = "Level 1 resolution:  dy = %g deg, %g min, %g sec = %g meters\n" \
        % (dy_deg,dy_min,dy_sec,dy_meters)
    if verbose:
        print(levtext)
    description = description + levtext

    amr_levels_max = rundata.amrdata.amr_levels_max
    refinement_ratios_y = rundata.amrdata.refinement_ratios_y
    num_ref_ratios = len(refinement_ratios_y)
    if amr_levels_max > num_ref_ratios+1:
        raise IOError("*** Too few refinement ratios specified for " \
            + "amr_levels_max = %i" % amr_levels_max)
    dy_levels = (num_ref_ratios+1) * [dy]
    for k,r in enumerate(refinement_ratios_y):
        level = k+2
        dy = dy_levels[k] / r
        dy_levels[k+1] = dy
        dy_meters = dy*111e3
        dy_deg,dy_min,dy_sec = deg2dms(dy)
        levtext = "Level %s resolution:  dy = %g deg, %g min, %g sec = %g meters  (refined by %i)\n" \
                % (level,dy_deg,dy_min,dy_sec,dy_meters,r)
        if verbose:
            print(levtext)
        description = description + levtext

    if verbose:
        print("Allowing maximum of %i levels" % amr_levels_max)

    elev = 0.
    if not combined:
        fname = 'Domain.kml'
    else:
        fname = fname_combined

    kml_text = kml_header(fname)

    mapping = {}
    mapping['x1'] = x1
    mapping['x2'] = x2
    mapping['y1'] = y1
    mapping['y2'] = y2
    mapping['elev'] = elev
    mapping['name'] = 'Computational Domain'
    mapping['desc'] = description
    mapping['color'] = "0000FF"  # red
    mapping['width'] = 2

    region_text = kml_region(mapping)
    kml_text = kml_text + region_text

    if not combined:
        kml_text = kml_text + kml_footer()
        kml_file = open(fname,'w')
        kml_file.write(kml_text)
        kml_file.close()
        if verbose:
            print("Created ",fname)

            

    regions = rundata.regiondata.regions
    if len(regions)==0 and verbose:
        print("No regions found in setrun.py")


    for rnum,region in enumerate(regions):
        if not combined:
            fname = 'Region_%s.kml' % str(rnum).zfill(2)
            kml_text = kml_header(fname)

        minlevel,maxlevel = region[0:2]
        t1,t2 = region[2:4]
        x1,x2,y1,y2 = region[4:]

        if verbose:
            print("Region %i: %10.6f  %10.6f  %10.6f  %10.6f" \
                    % (rnum,x1,x2,y1,y2))
            print("           minlevel = %i,  maxlevel = %i" \
                    % (minlevel,maxlevel) \
                    + "  t1 = %s,  t2 = %s" % (f2s(t1),f2s(t2)))
        mapping = {}
        mapping['minlevel'] = minlevel
        mapping['maxlevel'] = maxlevel
        mapping['t1'] = t1
        mapping['t2'] = t2
        mapping['x1'] = x1
        mapping['x2'] = x2
        mapping['y1'] = y1
        mapping['y2'] = y2
        mapping['elev'] = elev
        mapping['name'] = 'Region %i' % rnum
        description = "minlevel = %i, maxlevel = %i\n" % (minlevel,maxlevel) \
            + "  t1 = %s, t2 = %s\n" % (f2s(t1),f2s(t2)) \
            + "  x1 = %s, x2 = %s\n" % (f2s(x1),f2s(x2)) \
            + "  y1 = %s, y2 = %s\n\n" % (f2s(y1),f2s(y2))
        if len(dy_levels) >= minlevel:
            dy = dy_levels[minlevel-1]
            dy_deg,dy_min,dy_sec = deg2dms(dy)
            dy_meters = dy*111e3
            levtext = "Level %s resolution:  \ndy = %g deg, %g min, %g sec \n= %g meters\n" \
                    % (minlevel,dy_deg,dy_min,dy_sec,dy_meters)
            description = description + levtext
        if (maxlevel > minlevel) and (len(dy_levels) >= maxlevel):
            dy = dy_levels[maxlevel-1]
            dy_deg,dy_min,dy_sec = deg2dms(dy)
            dy_meters = dy*111e3
            levtext = "\nLevel %s resolution:  \ndy = %g deg, %g min, %g sec \n= %g meters\n" \
                    % (maxlevel,dy_deg,dy_min,dy_sec,dy_meters)
            description = description + levtext
        mapping['desc'] = description
        mapping['color'] = "FFFFFF"  # white
        mapping['width'] = 3

        region_text = kml_region(mapping)
        kml_text = kml_text + region_text
        if not combined:
            kml_text = kml_text + kml_footer()
            kml_file = open(fname,'w')
            kml_file.write(kml_text)
            kml_file.close()
            if verbose:
                print("Created ",fname)


    try:
        flagregions = rundata.flagregiondata.flagregions
    except:
        flagregions = []  # flagregions not yet in amrclaw

    if len(flagregions)==0 and verbose:
        print("No flagregions found in setrun.py")

    for rnum,flagregion in enumerate(flagregions):
        
        name = flagregion.name
        #print('+++ flagregion name = ',name)
        
        if not combined:
            if name == '':
                fname = 'FlagRegion_%s.kml' % str(rnum).zfill(2)
            else:
                fname = name + '.kml'
            kml_text = kml_header(fname)
            
        #if flagregion.spatial_region is None:
        #    flagregion.read_spatial_region()
        if flagregion.spatial_region_type == 1:
            x1,x2,y1,y2 = flagregion.spatial_region
        else:
            flagregion.read_spatial_region()
            x1,x2,y1,y2 = flagregion.spatial_region.bounding_box()
        minlevel = flagregion.minlevel
        maxlevel = flagregion.maxlevel

        if verbose:
            print("Region ", flagregion.name)

        mapping = {}
        mapping['minlevel'] = flagregion.minlevel
        mapping['maxlevel'] = flagregion.maxlevel
        mapping['t1'] = flagregion.t1
        mapping['t2'] = flagregion.t2
        mapping['x1'] = x1
        mapping['x2'] = x2
        mapping['y1'] = y1
        mapping['y2'] = y2
        mapping['elev'] = elev
        mapping['name'] = flagregion.name
        description = "minlevel = %i, maxlevel = %i\n" \
            % (flagregion.minlevel,flagregion.maxlevel) \
            + "  t1 = %s, t2 = %s\n" % (f2s(flagregion.t1),f2s(flagregion.t2)) \
            + "  Bounding box: \n" \
            + "  x1_bb = %s, x2_bb = %s\n" % (f2s(x1),f2s(x2)) \
            + "  y1_bb = %s, y2_bb = %s\n\n" % (f2s(y1),f2s(y2))
        if len(dy_levels) >= minlevel:
            dy = dy_levels[minlevel-1]
            dy_deg,dy_min,dy_sec = deg2dms(dy)
            dy_meters = dy*111e3
            levtext = "Level %s resolution:  \ndy = %g deg, %g min, %g sec \n= %g meters\n" \
                    % (minlevel,dy_deg,dy_min,dy_sec,dy_meters)
            description = description + levtext
        if (maxlevel > minlevel) and (len(dy_levels) >= maxlevel):
            dy = dy_levels[maxlevel-1]
            dy_deg,dy_min,dy_sec = deg2dms(dy)
            dy_meters = dy*111e3
            levtext = "\nLevel %s resolution:  \ndy = %g deg, %g min, %g sec \n= %g meters\n" \
                    % (maxlevel,dy_deg,dy_min,dy_sec,dy_meters)
            description = description + levtext
        mapping['desc'] = description
        mapping['color'] = "00FFFF"  # yellow
        mapping['width'] = 2
        
        if flagregion.spatial_region_type == 1:
            x1,x2,y1,y2 = flagregion.spatial_region
            x = [x1,x1,x2,x2,x1]
            y = [y1,y2,y2,y1,y1]
        else:
            x,y = flagregion.spatial_region.vertices()
            
        v = "\n"
        for j in range(len(x)):
            v = v + "%s,%s,%s\n" % (f2s(x[j]),f2s(y[j]),f2s(elev))
        v = v + "%s,%s,%s\n" % (f2s(x[0]),f2s(y[0]),f2s(elev))
        v.replace(' ','')
        
        region_text = kml_region(mapping, v)

        fname = flagregion.name + '.kml'
        region_text = kml_region(mapping, v)
        kml_text = kml_text + region_text
        
        if not combined:
            kml_text = kml_text + kml_footer()
            kml_file = open(fname,'w')
            kml_file.write(kml_text)
            kml_file.close()
            if verbose:
                print("Created ",fname)
        
    if combined:
        fname = fname_combined
        kml_text = kml_text + kml_footer()
        kml_file = open(fname,'w')
        kml_file.write(kml_text)
        kml_file.close()
        if verbose:
            print("Created ",fname)



def line2kml(xy,fname='line.kml',name='line',color='00FFFF',width=3,
             verbose=True):
    """
    Make a KML line with default color yellow.

    :Inputs:

     - *xy* a tuple ((x1,x2),(y1,y2)) (preferred) 
            or (x1,x2,y1,y2) (for backward compatibility)
     - *fname* (str) name of resulting kml file
     - *name* (str) name to appear on line on Google Earth
     - *color* (str) Color in format aabbggrr
     - *width* (str) line width
     - *verbose* (bool) - If *True*, print out info

    """
     
    if type(xy[0]) in [tuple,list]:
        x1,x2 = xy[0]
        y1,y2 = xy[1]
    else:
        x1,x2,y1,y2 = xy[0:]

    if verbose:
        print("Line:   %10.6f  %10.6f  %10.6f  %10.6f" % (x1,x2,y1,y2))

    elev = 0.
    kml_text = kml_header(fname)

    mapping = {}
    mapping['x1'] = x1
    mapping['x2'] = x2
    mapping['y1'] = y1
    mapping['y2'] = y2
    mapping['elev'] = elev
    mapping['name'] = name
    mapping['desc'] = "  x1 = %s, x2 = %s\n" % (f2s(x1),f2s(x2)) \
            + "  y1 = %s, y2 = %s" % (f2s(y1),f2s(y2))
    mapping['color'] = color
    mapping['width'] = width

    region_text = kml_line(mapping)

    kml_text = kml_text + region_text + kml_footer()
    kml_file = open(fname,'w')
    kml_file.write(kml_text)
    kml_file.close()
    if verbose:
        print("Created ",fname)


def box2kml(xy,fname=None,name='box',color='FF0000',width=3,verbose=True):
    """
    Make a KML box with default color blue.

    :Inputs:

     - *xy* a tuple ((x1,x2),(y1,y2)) (preferred) 
            or (x1,x2,y1,y2) (for backward compatibility)
     - *fname* (str) name of resulting kml file
     - *name* (str) name to appear in box on Google Earth
     - *color* (str) Color in format aabbggrr
     - *width* (str) line width
     - *verbose* (bool) - If *True*, print out info

    """

    if fname is None:
        fname = name + '.kml'

    if type(xy[0]) in [tuple,list]:
        x1,x2 = xy[0]
        y1,y2 = xy[1]
    else:
        x1,x2,y1,y2 = xy[0:]

    if verbose:
        print("Box:   %10.6f  %10.6f  %10.6f  %10.6f" % (x1,x2,y1,y2))

    elev = 0.
    kml_text = kml_header(fname)

    mapping = {}
    mapping['x1'] = x1
    mapping['x2'] = x2
    mapping['y1'] = y1
    mapping['y2'] = y2
    mapping['elev'] = elev
    mapping['name'] = name
    mapping['desc'] = "  x1 = %s, x2 = %s\n" % (f2s(x1),f2s(x2)) \
            + "  y1 = %s, y2 = %s" % (f2s(y1),f2s(y2))
    mapping['color'] = color
    mapping['width'] = width

    region_text = kml_region(mapping)

    kml_text = kml_text + region_text + kml_footer()
    kml_file = open(fname,'w')
    kml_file.write(kml_text)
    kml_file.close()
    if verbose:
        print("Created ",fname)


def quad2kml(xy,fname=None,name='quad',color='FF0000',width=3,verbose=True):
    """
    Make a KML quadrilateral with default color blue.

    :Inputs:

     - *xy* a tuple ((x1,x2,x3,x4),(y1,y2,y3,y4)) (preferred) 
            or (x1,x2,y1,y2,x3,y3,x4,y4) (for backward compatibility)
     - *fname* (str) name of resulting kml file
     - *name* (str) name to appear in box on Google Earth
     - *color* (str) Color in format aabbggrr
     - *width* (str) line width
     - *verbose* (bool) - If *True*, print out info

    """

    if fname is None:
        fname = name + '.kml'

    if type(xy[0]) in [tuple,list]:
        x1,x2,x3,x4 = xy[0]
        y1,y2,y3,y4 = xy[1]
    else:
        x1,y1,x2,y2,x3,y3,x4,y4 = xy[0:]

    if verbose:
        print("Quadrilateral:   %10.6f  %10.6f" % (x1,y1))
        print("                 %10.6f  %10.6f" % (x2,y2))
        print("                 %10.6f  %10.6f" % (x3,y3))
        print("                 %10.6f  %10.6f" % (x4,y4))

    elev = 0.
    kml_text = kml_header(fname)

    mapping = {}
    mapping['x1'] = x1
    mapping['x2'] = x2
    mapping['x3'] = x3
    mapping['x4'] = x4
    mapping['y1'] = y1
    mapping['y2'] = y2
    mapping['y3'] = y3
    mapping['y4'] = y4
    mapping['elev'] = elev
    mapping['name'] = name
    mapping['desc'] = "  x1 = %s, y1 = %s\n" % (f2s(x1),f2s(y1)) \
            + "  x2 = %s, y2 = %s" % (f2s(x2),f2s(y2)) \
            + "  x3 = %s, y3 = %s" % (f2s(x3),f2s(y3)) \
            + "  x4 = %s, y4 = %s" % (f2s(x4),f2s(y4))
    mapping['color'] = color
    mapping['width'] = 3

    region_text = kml_region(mapping)

    kml_text = kml_text + region_text + kml_footer()
    kml_file = open(fname,'w')
    kml_file.write(kml_text)
    kml_file.close()
    if verbose:
        print("Created ",fname)


def poly2kml(xy,fname=None,name='poly',color='00FF00', width=3,
             verbose=True, max_vertices_in_description=20):
    """
    Make a KML polygon with default color blue.

    :Inputs:

     - *xy* a tuple (x,y) where x and y are lists of vertices
     - *fname* (str) name of resulting kml file
     - *name* (str) name to appear in box on Google Earth
     - *color* (str) Color in format aabbggrr
     - *width* (str) line width
     - *verbose* (bool) - If *True*, print out info
     - *max_vertices_in_description* (int) - if more than this number
       of vertices, only list number in description box, not all vertices
    """

    if fname is None:
        fname = name + '.kml'

    x,y = xy

    if verbose:
        print("Creating kml for polygon with %i vertices" % len(x))
        if (len(x) <= max_vertices_in_description):
            for j in range(len(x)):
                print("             %10.6f  %10.6f" % (x[j],y[j]))            

    elev = 0.
    kml_text = kml_header(fname)

    mapping = {}
    mapping['x'] = x
    mapping['y'] = y
    mapping['elev'] = elev
    mapping['name'] = name
    d = "  Polygon with %i vertices" % len(x)
    if (len(x) <= max_vertices_in_description):
        d = "  x[0] = %s, y[0] = %s\n" % (f2s(x[0]),f2s(y[0])) 
        for j in range(1,len(x)):
            d = d + "  x[%i] = %s, y[%i] = %s\n" % (j,f2s(x[j]),j,f2s(y[j]))
    mapping['desc'] = d
    mapping['color'] = color
    mapping['width'] = width

    v = "\n"
    for j in range(len(x)):
        v = v + "%s,%s,%s\n" % (f2s(x[j]),f2s(y[j]),f2s(elev))
    v = v + "%s,%s,%s\n" % (f2s(x[0]),f2s(y[0]),f2s(elev))
    v.replace(' ','')
    
    region_text = kml_region(mapping, v)
    for j in range(1,len(x)):
        d = d + "  x[%i] = %s, y[%i] = %s" % (j,f2s(x[j]),j,f2s(y[j]))

    kml_text = kml_text + region_text + kml_footer()
    kml_file = open(fname,'w')
    kml_file.write(kml_text)
    kml_file.close()
    if verbose:
        print("Created ",fname)


def gauges2kml(rundata=None, fname='gauges.kml', verbose=True):

    """

    Create a KML marker for each gauge specified for a GeoClaw run.

    :Inputs:

      - *rundata* - an object of class *ClawRunData* or None

        If *rundata==None*, try to create based on executing function *setrun*
        from the `setrun.py` file in the current directory.

      - *fname* (str) - resulting kml file.

      - *verbose* (bool) - If *True*, print out info about each region found


    :Example:

        >>> from clawpack.geoclaw import kmltools
        >>> kmltools.gauges2kml()

    is equivalent to:

        >>> from clawpack.geoclaw import kmltools
        >>> from setrun import setrun
        >>> rundata = setrun()
        >>> kmltools.gauges2kml(rundata)

    By default this creates a file named *gauges.kml* that can be opened in
    Google Earth.

    """

    if rundata is None:
        try:
            import setrun
            reload(setrun)
            rundata = setrun.setrun()
        except:
            raise IOError("*** cannot execute setrun file")

    elev = 0.
    kml_text = kml_header(fname)


    gauges = rundata.gaugedata.gauges
    if len(gauges)==0 and verbose:
        print("No gauges found in setrun.py")


    for rnum,gauge in enumerate(gauges):
        t1,t2 = gauge[3:5]
        x1,y1 = gauge[1:3]
        gaugeno = gauge[0]
        if verbose:
            print("Gauge %i: %s, %s  \n" % (gaugeno,f2s(x1),f2s(y1)) \
                    + "  t1 = %s,  t2 = %s" % (f2s(t1),f2s(t2)))
        mapping = {}
        mapping['gaugeno'] = gaugeno
        mapping['t1'] = t1
        mapping['t2'] = t2
        mapping['x1'] = x1
        mapping['y1'] = y1
        mapping['elev'] = elev
        mapping['name'] = 'Gauge %i' % rnum
        description = "  t1 = %s, t2 = %s\n" % (f2s(t1),f2s(t2)) \
            + "  x1 = %s, y1 = %s\n" % (f2s(x1),f2s(y1))
        mapping['desc'] = description

        gauge_text = kml_gauge(mapping)
        kml_text = kml_text + gauge_text
    kml_text = kml_text + kml_footer()
    kml_file = open(fname,'w')
    kml_file.write(kml_text)
    kml_file.close()
    if verbose:
        print("Created ",fname)



def kml_header(name='GeoClaw kml file'):
    header = """<?xml version="1.0" encoding="UTF-8"?>
<kml xmlns="http://www.opengis.net/kml/2.2"
xmlns:gx="http://www.google.com/kml/ext/2.2">
<Document><name>%s</name>
""" % name
    return header

def kml_footer():
    footer = """
</Document>
</kml>
"""
    return footer


def kml_region(mapping, vertex_text=None):

    if vertex_text is None:
        if 'x3' in mapping:
            # quadrilateral with 4 corners specified
            vertex_text = """
{x1:.9f},{y1:.9f},{elev:.9f}
{x2:.9f},{y2:.9f},{elev:.9f}
{x3:.9f},{y3:.9f},{elev:.9f}
{x4:.9f},{y4:.9f},{elev:.9f}
{x1:.9f},{y1:.9f},{elev:.9f}
""".format(**mapping).replace(' ','')

        else:
            # rectangle with 2 corners specified
            vertex_text = """
{x1:.9f},{y1:.9f},{elev:.9f}
{x2:.9f},{y1:.9f},{elev:.9f}
{x2:.9f},{y2:.9f},{elev:.9f}
{x1:.9f},{y2:.9f},{elev:.9f}
{x1:.9f},{y1:.9f},{elev:.9f}
""".format(**mapping).replace(' ','')

    mapping['vertices'] = vertex_text
    if len(mapping['color'])==6:
        mapping['color'] = 'FF' + mapping['color']

    kml_text = """
<Style id="Path">
<LineStyle><color>{color:s}</color><width>{width:d}</width></LineStyle>
<PolyStyle><color>00000000</color></PolyStyle>
</Style>
<Placemark><name>{name:s}</name>
<description>{desc:s}</description>
<styleUrl>#Path</styleUrl>
<Polygon>
<tessellate>1</tessellate>
<altitudeMode>clampToGround</altitudeMode>
<outerBoundaryIs><LinearRing><coordinates>
{vertices:s}
</coordinates></LinearRing></outerBoundaryIs>
</Polygon>
</Placemark>
""".format(**mapping)

    return kml_text

def kml_line(mapping):

    if len(mapping['color'])==6:
        mapping['color'] = 'FF' + mapping['color']

    line_text = """
{x1:.9f},{y1:.9f},{elev:.9f}
{x2:.9f},{y2:.9f},{elev:.9f}
""".format(**mapping).replace(' ','')

    mapping['line'] = line_text
    kml_text = """
<Style id="Path">
<LineStyle><color>{color:s}</color><width>{width:d}</width></LineStyle>
<PolyStyle><color>00000000</color></PolyStyle>
</Style>
<Placemark><name>{name:s}</name>
<description>{desc:s}</description>
<styleUrl>#Path</styleUrl>
<LineString>
<tessellate>1</tessellate>
<altitudeMode>clampToGround</altitudeMode>
<coordinates>
{line:s}
</coordinates>
</LineString>
</Placemark>
""".format(**mapping)

    return kml_text

def kml_gauge(mapping):
    gauge_text = "{x1:.9f},{y1:.9f},{elev:.9f}".format(**mapping).replace(' ','')

    mapping['gauge'] = gauge_text

    kml_text = """
<Placemark><name>Gauge {gaugeno:d}</name>
<description>{desc:s}</description>
<styleUrl>#markerstyle</styleUrl>
<Point>
<coordinates>
{gauge:s}
</coordinates>
</Point>
</Placemark>
""".format(**mapping)

    return kml_text



def kml_timespan(t1,t2,event_time=None,tz=None,tscale=1):

    r"""
    Create time strings necessary for sliders in Google Earth.  The time
    span will cover time [t1,t2], with the start of the event given by
    event_time.

    [t1,t2]    : time span,

    event_time : Start of event in UTC :  [Y,M,D,H,M,S], e.g. [2010,2,27,3,34,0]
    tz         : time zone offset to UTC.  e.g. +3 for Chile; -9 for Japan.

    Time span element looks like ::

        <TimeSpan>
          <begin>2010-02-27T06:34:00+03:00</begin>
          <end>2010-02-27T07:04:00+03:00</end>
        </TimeSpan>

    As for how well this handles  Daylight  Savings time, here is what the documentation
    on the Python 'time' module has to say :

    "DST is Daylight Saving Time, an adjustment of the timezone by (usually) one hour
    during part of the year. DST rules are magic (determined by local law) and can
    change from year to year. The C library has a table containing the local rules
    (often it is read from a system file for flexibility) and is the only source of
    True Wisdom in this respect."

    """

    t1 = t1*tscale   # Time converted to seconds
    t2 = t2*tscale

    import time
    # to adjust time from UTC to time in event locale.
    if event_time == None:
        # Use local time.
        starttime = time.mktime(time.localtime())  # seconds UTC
        tz_offset = time.timezone/3600.0   # in seconds
    else:
        ev = tuple(event_time) + (0,0,0)   # Extend to 9 tuple; no DST
        # mktime returns time in seconds + timezone offset, i.e. seconds UTC
        # Subtract out the timezone offset here, since it will get added back
        # in when we do gmtime(starttime + ...) below.
        starttime = time.mktime(ev) - time.timezone
        if tz is None:
            print("===> Time zone offset not defined;  assuming zero offset. " \
                "Set plotdata.kml_tz_offset to define an offset (in hours) from "\
                "UTC (positive west of UTC; negative east of UTC)")
            tz = 0

        tz_offset = tz

    if (tz_offset == None):
        tzstr = "Z"  # no offset; could also just set to "+00:00"
    else:
        # Google Earth will show time slider time in local time, where
        # local + offset = UTC.
        tz_offset = tz_offset*3600.    # Offset in seconds
        tz = time.gmtime(abs(tz_offset))
        if (tz_offset > 0):
            tzstr = time.strftime("+%H:%M",tz)  # Time to UTC
        else:
            tzstr = time.strftime("-%H:%M",tz)

    # Get time strings for start and end of time span
    gbegin = time.gmtime(starttime + t1)
    timestrbegin = "%s%s" % (time.strftime("%Y-%m-%dT%H:%M:%S", gbegin),tzstr)

    gend = time.gmtime(starttime + t2)
    timestrend = "%s%s" % (time.strftime("%Y-%m-%dT%H:%M:%S", gend),tzstr)

    return timestrbegin,timestrend

def topo2kml(topo_file_name, topo_type, color='00FF00'):       
    """
    Create a kml file putting a box around the region covered by a topofile.
    Color is green by default.
    """

    import os
    from clawpack.geoclaw import topotools
    topo = topotools.Topography(topo_file_name, topo_type=topo_type)
    topo.read_header()
    xy = topo.extent
    name = os.path.splitext(os.path.split(topo_file_name)[-1])[0]
    file_name = '%s.kml' % name
    box2kml(xy, file_name, name, color)

def dtopo2kml(dtopo_file_name, dtopo_type, color='8888FF'):       
    """
    Create a kml file putting a box around the region covered by a dtopofile.
    Color is pink by default.
    """

    import os
    from clawpack.geoclaw import dtopotools
    dtopo = dtopotools.DTopography()
    dtopo.read(dtopo_file_name, dtopo_type)
    x1 = dtopo.x.min()
    x2 = dtopo.x.max()
    y1 = dtopo.y.min()
    y2 = dtopo.y.max()
    xy = (x1,x2,y1,y2)
    name = os.path.splitext(os.path.split(dtopo_file_name)[-1])[0]
    file_name = '%s.kml' % name
    box2kml(xy, file_name, name, color)
        
        

def fgmax2kml(rundata=None,fname='fgmax_grids.kml',verbose=True,combined=False):

    """
    Create a KML box for each fgmax grid specified for a GeoClaw run.

    :Inputs:

      - *rundata* - an object of class *ClawRunData* or None

        If *rundata==None*, try to create based on executing function *setrun*
        from the `setrun.py` file in the current directory.

      - *fname* (str) - resulting kml file.

      - *verbose* (bool) - If *True*, print out info about each region found

      - *combined* (bool) - If *True*, combine into single kml file with
        name given by *fname*.  NOT YET IMPLEMENTED.
        If False, *fname* is ignored and individual files are created for
        each fgmax grid.

    """

    from numpy import cos,pi,floor

    if rundata is None:
        try:
            import setrun
            reload(setrun)
            rundata = setrun.setrun()
        except:
            raise IOError("*** cannot execute setrun file")

    if combined:
        fname_combined = 'fgmax_grids.kml'
        print('*** combined fgmax kml files not yet supported')
        print('    making a kml file for each fgmax grid')
    
    fgmax_grids = rundata.fgmax_data.fgmax_grids

    for fg in fgmax_grids:
        fname_root = 'fgmax%s' % str(fg.fgno).zfill(4)
        kml_file = fname_root + '.kml'
        if fg.point_style==1:
            xy = ([fg.x1,fg.x2], [fg.y1,fg.y2])
            line2kml(xy,kml_file, fname_root, color='8888FF', width=2)
        if fg.point_style==2:
            xy = ([fg.x1,fg.x2], [fg.y1,fg.y2])
            box2kml(xy, kml_file, fname_root, color='8888FF')
        elif fg.point_style==3:
            xy = ([fg.x1,fg.x2,fg.x3,fg.x4], 
                  [fg.y1,fg.y2,fg.y3,fg.y4])
            poly2kml(xy, kml_file, fname_root, color='8888FF')
        else:
            print('fgmax2kml not yet implemented for point_style = %i' \
                  % fg.point_style)


def fgout2kml(rundata=None,fname='fgout_grids.kml',verbose=True,combined=False):

    """
    Create a KML box for each fgout grid specified for a GeoClaw run.

    :Inputs:

      - *rundata* - an object of class *ClawRunData* or None

        If *rundata==None*, try to create based on executing function *setrun*
        from the `setrun.py` file in the current directory.

      - *fname* (str) - resulting kml file.

      - *verbose* (bool) - If *True*, print out info about each region found

      - *combined* (bool) - If *True*, combine into single kml file with
        name given by *fname*.  NOT YET IMPLEMENTED.
        If False, *fname* is ignored and individual files are created for
        each fgout grid.

    """

    from numpy import cos,pi,floor

    if rundata is None:
        try:
            import setrun
            reload(setrun)
            rundata = setrun.setrun()
        except:
            raise IOError("*** cannot execute setrun file")

    if combined:
        fname_combined = 'fgout_grids.kml'
        print('*** combined fgout kml files not yet supported')
        print('    making a kml file for each fgout grid')
    
    fgout_grids = rundata.fgout_data.fgout_grids

    for fg in fgout_grids:
        fname_root = 'fgout%s' % str(fg.fgno).zfill(4)
        kml_file = fname_root + '.kml'
        xy = ([fg.x1,fg.x2], [fg.y1,fg.y2])
        box2kml(xy, kml_file, fname_root, color='8888FF')


def make_input_data_kmls(rundata=None, combined=False):
    """
    Produce kml files for the computational domain, all gauges and regions 
    specified, and all topo and dtopo files specified in rundata.
    This can be used, e.g. by adding the lines 

        from clawpack.geoclaw import kmltools
        kmltools.make_input_data_kmls(rundata)

    to the end of a `setrun.py` file so that `make data` will generate all
    kml files in addition to the `*.data` files.
    
    Or set *rundata==None*, in which case it will try to generate rundata
    based on executing function *setrun*
    from the `setrun.py` file in the current directory.
    """
    
    import os
    from clawpack.geoclaw import topotools, dtopotools
    
    if rundata is None:
        try:
            import setrun
            reload(setrun)
            rundata = setrun.setrun()
        except:
            raise IOError("*** cannot execute setrun file")

    regions2kml(rundata, combined=combined)
    gauges2kml(rundata)
    fgmax2kml(rundata)
    fgout2kml(rundata)

    topofiles = rundata.topo_data.topofiles
    for f in topofiles:
        topo_file_name = f[-1]
        topo_type = f[0]
        topo2kml(topo_file_name, topo_type)
        
    dtopofiles = rundata.dtopo_data.dtopofiles
    for f in dtopofiles:
        dtopo_file_name = f[-1]
        dtopo_type = f[0]
        dtopo2kml(dtopo_file_name, dtopo_type)
        

def pcolorcells_for_kml(X, Y, Z, png_filename=None, dpc=2, max_inches=15., 
                        verbose=True, **kwargs):
    
    """
    Wraps pcolormesh in a way that a png file is created that can be viewed
    on Google Earth with proper alignment and with sharp grid cell edges.
    Works if X,Y are cell centers or edges, and X,Y can be 2d or 1d arrays. 
    
    X,Y,Z is the data to be plotted.  It is assumed to be finite volume data
    where Z[i,j] is a constant value over a grid cell.

    Internally x,y are defined as 1d arrays since it is assumed the 
    grids are Cartesian.
    
    If the length of the 1d arrays x and y match the dimensions of Z then 
    these are assumed to be cell center values. In this case the arrays 
    are expanded by one to obtain x_edge, y_edge as edge values,
    as needed for proper alignment.  
    
    If the length of x,y is already one greater than the corresponding 
    dimension of Z, then it is assumed that these are already edge values.

    If png_filename is not None then a png file is written with appropriate dpi.

    dpc is the desired "dots per cell", how many pixels to allot to each
    to each grid cell.  This should be an integer to avoid interpolation
    between cells that smears out the cell boundaries in the png file.
    Increasing this will give sharper boundaries but also larger files that
    load more slowly.
    
    max_inches is the desired size of the longer edge of the figure created.
    This value is not very important unless you want to view the png file
    on a screen outside of Google Earth.  Internally the dimensions of the
    figure `x_inches` and `y_inches` are determined to be consistent
    with the value `dpc` specified and a reasonable value of `dpi` for the
    png file, as described below.

    Internally the value `dpi` (dots per inch) for the png file is
    determined so that it is at least 16 and so that:
        dpi * x_inches = dcp * x_cells
        dpi * y_inches = dcp * y_cells
    where `x_cells`, `y_cells` are the number of cells in each direction.

    `kwargs` are passed to `pcolormesh`, e.g. `cmap` and `norm` are
    generally specified.

    This function returns `fig, ax, png_extent, kml_dpi` so the user can further
    annotate the figure befor saving it as a png file, which should then 
    be done with:
        plt.savefig(png_filename, transparent=True, dpi=kml_dpi)
    The `png_extent` is needed in construcing a kml file to display the
    png file on Google Earth, e.g. using the function `png2kml` in this
    module.
    """

    from matplotlib import pyplot as plt
    import numpy as np
    
    # If X is 2d extract proper 1d slice:
    if X.ndim == 1:
        x = X
    elif X.ndim == 2:
        if X[0,0] == X[0,1]:
            x = X[:,0]
        else:
            x = X[0,:]
            
    # If Y is 2d extract proper 1d slice:
    if Y.ndim == 1:
        y = Y
    elif Y.ndim == 2:
        if Y[0,0] == Y[0,1]:
            y = Y[:,0]
        else:
            y = Y[0,:]                    

    dx = x[1]-x[0]
    dy = y[1]-y[0]
    if len(x) == Z.shape[1]:
        # cell centers, so xedge should be expanded by dx/2 on each end:
        xedge = np.arange(x[0]-0.5*dx, x[-1]+dx, dx)
    elif len(x) == Z.shape[1]+1:
        # assume x already contains edge values
        xedge = x
    else:
        raise ValueError('x has unexpected length')

    if len(y) == Z.shape[0]:
        # cell centers, so xedge should be expanded by dx/2 on each end:
        yedge = np.arange(y[0]-0.5*dy, y[-1]+dy, dy)
    elif len(y) == Z.shape[0]+1:
        # assume x already contains edge values
        yedge = y
    else:
        raise ValueError('y has unexpected length')
        

    x1 = xedge[0];  x2 = xedge[-1]
    y1 = yedge[0];  y2 = yedge[-1]

    # Number of grid cells:
    x_cells = int(round((x2-x1)/dx))
    y_cells = int(round((y2-y1)/dy))
    max_cells = max(x_cells, y_cells)
            
    dots_per_cell = dpc
    max_dots = dots_per_cell * max_cells

    # determine dots per inch for png file, minimum 16:
    kml_dpi = max(int(round(max_dots / max_inches)), 16)
    dots_x = x_cells * dots_per_cell
    dots_y = y_cells * dots_per_cell

    # determine dimensions for figsize:
    x_inches = dots_x / kml_dpi
    y_inches = dots_y / kml_dpi
    dpc_x = kml_dpi * x_inches / x_cells
    dpc_y = kml_dpi * y_inches / y_cells

    if verbose:
        print('Using kml_dpi = %i,figure size %.6f by %.6f inches' \
                % (kml_dpi,x_inches,y_inches))
        print('Figure has %i by %i grid cells of uniform color' \
                % (x_cells, y_cells))
        print('Dots per cell in x: %.6f,  in y: %.6f' % (dpc_x,dpc_y))
        print('       These should be integers')
    
        
    # Create figure of appropriate size and pcolormesh plot
    # with no margins, ticks, or labels:

    fig = plt.figure(figsize=(x_inches,y_inches))
    ax = plt.axes()
    plt.axis('off')
    pc = plt.pcolormesh(xedge, yedge, Z, **kwargs)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_frame_on(False)
    fig.set_size_inches(x_inches,y_inches)
    plt.subplots_adjust(top = 1, bottom = 0, right = 1, left = 0, 
                    hspace = 0, wspace = 0)
    plt.margins(0,0)
    
    if png_filename is not None:
        plt.savefig(png_filename, transparent=True, dpi=kml_dpi)
        if verbose:
            print('Created ',png_filename)

    png_extent = [xedge[0], xedge[-1], yedge[0], yedge[-1]]
    return fig, ax, png_extent, kml_dpi
        


def kml_png(mapping):
    """
    Create text for a png file overlay
    """

    kml_text = """
<GroundOverlay>
<name>{name:s}</name>
<visibility>1</visibility>
<gx:altitudeMode> clampToSeaFloor </gx:altitudeMode>
<Icon>
  <href>{png_file:s}</href>
</Icon>
<LatLonBox>
  <north>{y2:.9f}</north>
  <south>{y1:.9f}</south>
  <east>{x2:.9f}</east>
  <west>{x1:.9f}</west>
</LatLonBox>
</GroundOverlay>
""".format(**mapping)

    return kml_text
    
def kml_cb(mapping):
    """
    Create text for a colorbar png file overlay
    """

    kml_text = """
<ScreenOverlay>
  <name>{name:s}</name>
  <Icon>
    <href>{cb_file:s}</href>
  </Icon>
  <overlayXY x="{xfrac:.4f}" xunits="fraction" y="{yfrac:.4f}" yunits="fraction"/>
  <screenXY x="{xfrac:.4f}" xunits="fraction" y="{yfrac:.4f}" yunits="fraction"/>
</ScreenOverlay>
""".format(**mapping)

    return kml_text

radio_style_text = """
<Style id="folderStyle">
<ListStyle>
<listItemType>radioFolder</listItemType>
</ListStyle>
</Style>
<styleUrl>#folderStyle</styleUrl>
"""



def png2kml(extent, png_files, png_names=None, name='png_files', fname=None, 
            radio_style=False, cb_files=None, cb_names=None, 
            cb_xfracs=None, cb_yfracs=None, verbose=True):
    """
    Create a kml file `fname` linking overlays for each png file in `png_files`.
    
    `extent` is [x1,x2,y1,y2] specifying where image should be overlaid.
    
    `png_names`, if present, will give the name for each image for the
    Google Earth menu.
    
    If `radio_style` is True, set radio buttons so only one can be shown at a 
    time, useful for combining plots of different quantities in same file.
    """
    
    import os
    
    if fname is None:
        fname = name + '.kml'


    x1,x2,y1,y2 = extent

    if verbose:
        print("Extent:   %10.6f  %10.6f  %10.6f  %10.6f" % (x1,x2,y1,y2))

    kml_text = kml_header(fname) + \
        "<name>%s</name>\n" % name + "<open>1</open>\n"
      

    mapping = {}
    mapping['x1'] = x1
    mapping['x2'] = x2
    mapping['y1'] = y1
    mapping['y2'] = y2
    
    for k,png_file in enumerate(png_files):
        
        mapping['png_file'] = png_file
        
        try:
            mapping['name'] = png_names[k]
        except:
            mapping['name'] =  'No name'

        kml_text = kml_text + kml_png(mapping)
        
    if radio_style:
        kml_text = kml_text + radio_style_text
        
    if cb_files:
        # colorbars
        for k,cb_file in enumerate(cb_files):
            
            mapping['cb_file'] = cb_file
            try:
                mapping['name'] = cb_names[k]
            except:
                mapping['name'] =  'Colorbar'
                
            try:
                mapping['xfrac'] = cb_xfracs[k]
            except:
                mapping['xfrac'] = 0.025 + k*0.075
            try:
                mapping['yfrac'] = cb_yfracs[k]
            except:
                mapping['yfrac'] = 0.05
                
            kml_text = kml_text + kml_cb(mapping)
        
    kml_text = kml_text + kml_footer()
    kml_file = open(fname,'w')
    kml_file.write(kml_text)
    kml_file.close()
    
    if verbose:
        print("Created ",fname)


def kml_build_colorbar(cb_filename, cmap, cmin=None, cmax=None, 
                       norm=None, label=None, title=None, extend='neither',
                       close_figs=True):

    """
    Make a png file with a colorbar corresponding to cmap, norm.
    cmin, cmax are used only if nrm is not provided.
    """
    
    import matplotlib.pyplot as plt
    import matplotlib as mpl
    import numpy

    fig = plt.figure(figsize=(1.2,3))
    ax1 = fig.add_axes([0.3, 0.075, 0.20, 0.80])
    tick = ax1.yaxis.get_major_ticks()
    plt.tick_params(axis='y', which='major', labelsize=8)

    if norm is None:
        norm = mpl.colors.Normalize(vmin=cmin,vmax=cmax)

    cb1 = mpl.colorbar.ColorbarBase(ax1,cmap=cmap,
                                    norm=norm,
                                    extend=extend,
                                    orientation='vertical')

    # make sure ticks appear at lower and upper limits of scale:
    cbticks = cb1.get_ticks()
    cbticks = numpy.hstack([norm.vmin, cbticks, norm.vmax])
    # remove 2nd and/or next-to-last tick if they are cramped:
    if cbticks[1]-cbticks[0] < 0.25*(cbticks[2]-cbticks[1]):
        cbticks = numpy.hstack((cbticks[:1], cbticks[2:]))
    if cbticks[-1]-cbticks[-2] < 0.25*(cbticks[-2]-cbticks[-3]):
        cbticks = numpy.hstack((cbticks[:-2], cbticks[-1:]))
    cb1.set_ticks(cbticks)

    if label:
        cb1.set_label(label)
    if title:
        ax1.set_title(title)
        
        
    # This is called from plotpages, in <plotdir>.
    plt.savefig(cb_filename,transparent=False)
    
    if close_figs:
        plt.close(fig)


def topo2kmz(topo, zlim=(-20,20), mask_outside_zlim=True, sea_level=0., 
             name='topo', force_dry=None, close_figs=True):

    """
    Create kmz file showing onshore and offshore topography as separate layers.
    Currently not very flexible but is useful in quickly creating kmz files
    to open on Google Earth showing the onshore topography and offshore
    bathymetry as two separate layers.  Constant colors on rectangular grid
    cells should be properly rendered to ease exploration of DEMs.
    A colorbar png file is also created and included in the kmz file. 

    :Input:
     - *topo* should be a topotools.Topography object,
     - *zlim* is the elevation z limits for choosing the color map
     - *mask_outside_zlim* if True, suppress plotting outsize `zlim`
     - *sea_level* is the break between water and land colors
     - *name* is used in the kml menu and file name
     - *force_dry* is currently not used 
     - *close_figs* to close the pyplot figures after making png files

    :Future:
    If `force_dry` is an array of the same shape as `topo.Z` then another png
    and kml file are created for land that is below `sea_level` but where
    `force_dry = True`.

    """
    
    import os, glob
    import numpy
    from numpy import ma
    from clawpack.visclaw import colormaps
    import zipfile
    import matplotlib.pyplot as plt
    
    assert force_dry is None, 'force_dry not yet implemented'
    
    cmap_land = colormaps.make_colormap({ 0.0:[0.1,0.4,0.0],
                                         0.25:[0.0,1.0,0.0],
                                          0.5:[0.8,1.0,0.5],
                                          1.0:[0.8,0.5,0.2]})

    cmap_water = colormaps.make_colormap({ 0.0:[0,0,1], 1.:[.8,.8,1]})

    cmap_topo, norm_topo = colormaps.add_colormaps((cmap_land, cmap_water),
                                         data_limits=(zlim[0],zlim[1]),
                                         data_break=sea_level)

    cmap_force_dry = colormaps.make_colormap({ 0.0:[1.0,0.7,0.7], 1.:[1.0,0.7,0.7]})
    cmap_dry, norm_dry = colormaps.add_colormaps((cmap_land, cmap_force_dry),
                                         data_limits=(zlim[0],zlim[1]),
                                         data_break=sea_level)
    
    if mask_outside_zlim:
        Z = ma.masked_where(numpy.logical_or(topo.Z<zlim[0], topo.Z>zlim[1]), topo.Z)
        cbar_extend = 'neither'
    else:
        Z = topo.Z
        cbar_extend = 'both'
    
    kml_dir = 'kmlfiles_%s' % name
    os.system('mkdir -p %s' % kml_dir)
    print('Will put png and kml files in %s' % kml_dir)
    
    Z_land = ma.masked_where(Z<sea_level, Z)
    png_filename = '%s/%s_land.png' % (kml_dir, name)
    fig,ax,png_extent,kml_dpi = pcolorcells_for_kml(topo.X, topo.Y, 
                                         Z_land, png_filename=png_filename,
                                         dpc=2, cmap=cmap_topo, norm=norm_topo)
    if close_figs:
        plt.close(fig)

    Z_water = ma.masked_where(Z>=sea_level, Z)
    png_filename = '%s/%s_water.png' % (kml_dir, name)
    fig,ax,png_extent,kml_dpi = pcolorcells_for_kml(topo.X, topo.Y, 
                                         Z_water, png_filename=png_filename,
                                         dpc=2, cmap=cmap_topo, norm=norm_topo)

    if close_figs:
        plt.close(fig)

    kml_build_colorbar('%s/colorbar.png' % kml_dir, cmap_topo, 
                                cmin=zlim[0], cmax=zlim[1], label='meters', 
                                title='topo', extend=cbar_extend,
                                close_figs=close_figs)

        
    png_files=['%s_water.png' % name, '%s_land.png' % name]
    png_names=['%s_water' % name,'%s_land' % name]
    cb_files = ['colorbar.png']
    cb_names = ['colorbar_topo']

    name = '%s_topo' % name
    fname = os.path.join(kml_dir, name+'.kml')
    png2kml(png_extent, png_files=png_files, png_names=png_names, 
                     name=name, fname=fname,
                     radio_style=False,
                     cb_files=cb_files, cb_names=cb_names)

    savedir = os.getcwd()
    os.chdir(kml_dir)
    files = glob.glob('*.kml') + glob.glob('*.png')
    print('kmz file will include:')
    for file in files:
        print('    %s' % os.path.split(file)[-1])

    fname_kmz = 'topo_%s.kmz' % name
    with zipfile.ZipFile(fname_kmz, 'w') as zip:
        for file in files:
            zip.write(file) 
        print('Created %s' % os.path.abspath(fname_kmz))
    os.chdir(savedir)

def transect2kml(xtrans, ytrans, fname='transect.kml'):
    """
    Create a kml file for points with long,lat specified by xtrans,ytrans.
    Label each point with number and coordinates.
    Adjust longitudes in coordinates so they are all between -180 and 180
    to display properly in Google Earth (but labels are original values).
    """

    with open(fname,'w') as kml_file:
           
        kml_file.write("""<?xml version="1.0" encoding="UTF-8"?>
        <kml xmlns="http://www.opengis.net/kml/2.2"
        xmlns:gx="http://www.google.com/kml/ext/2.2">
        <Document><name>transect_points</name>
        
        <Style id="Red">
        <IconStyle><color>FF0000FF</color></IconStyle>
        </Style>
        
        <Style id="Yellow">
        <IconStyle><color>FF00FFFF</color></IconStyle>
        </Style>
        """)
            
        for k in range(len(xtrans)):
            x = xtrans[k]
            y = ytrans[k]
            if x < -180:
                xge = x+360
            elif x > 180:
                xge = x-360
            else:
                xge = x
            style_id = 'Red'
            name = 'Transect point %i' % k
            kml_file.write("""
            <Placemark><name>%s</name>
            <description>%.5f, %.5f</description>
            <styleUrl>#%s</styleUrl>
            <Point>
            <coordinates>
            %.9f, %.9f, 0.0
            </coordinates>
            </Point>
            </Placemark>
            """ % (name,x,y,style_id,xge,y))
        
        kml_file.write("\n</Document>\n</kml>")
            
    print('Created ', fname)
    



r"""
Execute nosetests in all subdirectories, to run a series of quick
regression tests.

Sends output and result/errors to separate files to simplify checking
results and looking for errors.
"""

from __future__ import absolute_import
import os
import glob

import clawpack.clawutil.test
import clawpack.pyclaw.util

# Clean library files whenever this module is used
if "CLAW" in os.environ:
    CLAW = os.environ["CLAW"]
else:
    raise ValueError("Need to set CLAW environment variable.")

for lib_path in [os.path.join(CLAW,"amrclaw","src","2d"),
                 os.path.join(CLAW,"amrclaw","src","3d")]:
    for path in glob.glob(os.path.join(lib_path,"*.o")):
        os.remove(path)
    for path in glob.glob(os.path.join(lib_path,"*.mod")):
        os.remove(path)


class AMRClawRegressionTest(clawpack.clawutil.test.ClawpackRegressionTest):

    r"""Base AMRClaw regression test setup derived from ClawpackRegressionTest

    """

    __doc__ += clawpack.pyclaw.util.add_parent_doc(
                                  clawpack.clawutil.test.ClawpackRegressionTest)


    def build_executable(self, executable_name="xamr"):
        r"""Build executable by running `make .exe` in test directory.

        Moves the resulting executable to the temporary directory.


        """

        super(AMRClawRegressionTest, self).build_executable(
                                                executable_name=executable_name)



def test_shallow_sphere():
    """Test solution of shallow water equations on the sphere."""
    from . import Rossby_wave

    def verify_shallow_sphere(claw):
        import os
        import numpy as np
        from clawpack.pyclaw.util import check_diff

        test_q = claw.frames[-1].state.get_q_global()
        test_height = test_q[0,:,:]

        thisdir = os.path.dirname(__file__)
        data_filename='swsphere_height.txt'
        expected_height = np.loadtxt(os.path.join(thisdir,data_filename))
        test_err = np.linalg.norm(expected_height-test_height)
        expected_err = 0
        return check_diff(expected_err, test_err, abstol=1e-4)

    from clawpack.pyclaw.util import test_app
    kwargs = {}   
    kwargs['disable_output']= True
    return test_app(Rossby_wave.setup,
                    verify_shallow_sphere,
                    kwargs)




import os as _os

_subpackages = {
    'clawutil':   'clawutil/src/python',
    'riemann':    'riemann',
    'visclaw':    'visclaw/src/python',
    'pyclaw':     'pyclaw/src',
    'petclaw':    'pyclaw/src',
    'forestclaw': 'pyclaw/src',
    'classic':    'classic/src/python',
    'amrclaw':    'amrclaw/src/python',
    'geoclaw':    'geoclaw/src/python',
    'dclaw':      'dclaw/src/python',
}

__all__ = list(_subpackages.keys())

_init = _os.path.abspath(__file__)
_root = _os.path.dirname(_os.path.dirname(_init))
_path = (
    _os.path.join(_root, *_sdir.split('/'))
    for _sdir in set(_subpackages.values())
)
__path__.extend(filter(_os.path.isdir, _path))
del _os, _init, _root, _path

__version__ = '5.11.0'   # must also be changed in setup.py



#!/usr/bin/env python

def configuration(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('pyclaw', parent_package, top_path)
    config.add_data_files('log.config')
    config.add_subpackage('classic')
    config.add_subpackage('sharpclaw')
    config.add_subpackage('fileio')
    config.add_subpackage('limiters')
    config.add_subpackage('examples', subpackage_path='pyclaw/examples')
    config.add_subpackage('tests')
    return config


if __name__ == '__main__':
    from numpy.distutils.core import setup
    setup(**configuration(top_path='').todict())



"""Additions for ForestClaw support"""

from __future__ import absolute_import
import os
import logging
import logging.config

# Default logging configuration file
_DEFAULT_LOG_CONFIG_PATH = os.path.join(os.path.dirname(__file__),
                                        '../pyclaw/log.config')
del os

# Setup loggers
logging.config.fileConfig(_DEFAULT_LOG_CONFIG_PATH)

__all__ = []

# Module imports - Note the only difference here is the geometry module
__all__.extend(['Controller', 'Dimension', 'Patch', 'Domain', 'Solution',
                'State', 'CFL'])
from clawpack.pyclaw.controller import Controller
from clawpack.pyclaw.solution import Solution
from clawpack.pyclaw.state import State
from clawpack.pyclaw.cfl import CFL
from clawpack.pyclaw.geometry import Dimension
from clawpack.pyclaw.geometry import Domain
from .geometry import Patch



#!/usr/bin/env python
# encoding: utf-8

"""Output package for Pyclaw"""

import logging
logger = logging.getLogger('pyclaw.fileio')

from . import ascii
__all__ = ['ascii.read', 'ascii.write']

from . import binary
__all__ += ['binary.read']

# Check for HDF 5 support
try:
    import h5py
    from . import hdf5
    __all__ += ['hdf5.read', 'hdf5.write']
except ImportError:
    logger.debug("No hdf5 support found.")

# Check for netcdf support
try:
    import netCDF4
    from . import netcdf
    __all__ += ['netcdf.read', 'netcdf.write']
except(ImportError):
    logger.debug("No netcdf4 support found.")



from __future__ import absolute_import
import clawpack.pyclaw

class State(clawpack.pyclaw.State):
    """Parallel State class"""

    __doc__ += clawpack.pyclaw.util.add_parent_doc(clawpack.pyclaw.state)

    @property
    def num_eqn(self):
        r"""(int) - Number of unknowns (components of q)"""
        if self.q_da is None:
            raise Exception('state.num_eqn has not been set.')
        else: return self.q_da.dof

    @property
    def num_aux(self):
        r"""(int) - Number of auxiliary fields"""
        if self.aux_da is None: return 0
        else: return self.aux_da.dof

    @property
    def mp(self):
        r"""(int) - Number of derived quantities (components of p)"""
        if self._p_da is None:
            raise Exception('state.mp has not been set.')
        else: return self._p_da.dof
    @mp.setter
    def mp(self,mp):
        if self._p_da is not None:
            raise Exception('You cannot change state.mp after p is initialized.')
        else:
            self._p_da = self._create_DA(mp)
            self.gpVec = self._p_da.createGlobalVector()

    @property
    def mF(self):
        r"""(int) - Number of derived quantities (components of p)"""
        if self._F_da is None:
            raise Exception('state.mF has not been set.')
        else: return self._F_da.dof
    @mF.setter
    def mF(self,mF):
        if self._F_da is not None:
            raise Exception('You cannot change state.mp after p is initialized.')
        else:
            self._F_da = self._create_DA(mF)
            self.gFVec = self._F_da.createGlobalVector()

    @property
    def q(self):
        r"""
        Array of solution values.
        """
        shape = self.grid.num_cells
        shape.insert(0,self.num_eqn)
        return self.gqVec.getArray().reshape(shape, order = 'F')
    @q.setter
    def q(self,val):
        self.gqVec.setArray(val.reshape([-1], order = 'F'))

    @property
    def p(self):
        r"""
        Array containing values of derived quantities for output.
        """
        if self._p_da is None: return 0
        shape = self.grid.num_cells
        shape.insert(0,self.mp)
        p=self.gpVec.getArray().reshape(shape, order = 'F')
        return p
    @p.setter
    def p(self,val):
        mp = val.shape[0]
        if self.gpVec is None: self.init_p_da(mp)
        self.gpVec.setArray(val.reshape([-1], order = 'F'))

    @property
    def F(self):
        r"""
        Array containing pointwise values (densities) of output functionals.
        This is just used as temporary workspace before summing.
        """
        if self._F_da is None: return 0
        shape = self.grid.num_cells
        shape.insert(0,self.mF)
        F=self.gFVec.getArray().reshape(shape, order = 'F')
        return F
    @F.setter
    def fset(self,val):
        mF = val.shape[0]
        if self.gFVec is None: self.init_F_da(mF)
        self.gFVec.setArray(val.reshape([-1], order = 'F'))

    @property
    def aux(self):
        """
        We never communicate aux values; every processor should set its own ghost cell
        values for the aux array.  The global aux vector is used only for outputting
        the aux values to file; everywhere else we use the local vector.
        """
        if self.aux_da is None: return None
        shape = self.grid.num_cells
        shape.insert(0,self.num_aux)
        aux=self.gauxVec.getArray().reshape(shape, order = 'F')
        return aux
    @aux.setter
    def aux(self,val):
        # It would be nice to make this work also for parallel
        # loading from a file.
        if self.aux_da is None: 
            num_aux=val.shape[0]
            self._init_aux_da(num_aux)
        self.gauxVec.setArray(val.reshape([-1], order = 'F'))
    @property
    def num_dim(self):
        return self.patch.num_dim


    def __init__(self,geom,num_eqn,num_aux=0):
        r"""
        Here we don't call super because q and aux must be properties in PetClaw
        but should not be properties in PyClaw.

        :attributes:
        patch - The patch this state lives on
        """

        from clawpack.pyclaw import geometry
        if isinstance(geom,geometry.Patch):
            self.patch = geom
        elif isinstance(geom,geometry.Domain):
            self.patch = geom.patches[0]
        else:
            raise Exception("""A PetClaw State object must be initialized with
                             a PetClaw Patch or Domain object.""")

        self.aux_da = None
        self.q_da = None

        self._p_da = None
        self.gpVec = None

        self._F_da = None
        self.gFVec = None

        # ========== Attribute Definitions ===================================
        self.problem_data = {}
        r"""(dict) - Dictionary of global values for this patch, 
            ``default = {}``"""
        self.t=0.
        r"""(float) - Current time represented on this patch, 
            ``default = 0.0``"""
        self.index_capa = -1
        self.keep_gauges = False
        r"""(bool) - Keep gauge values in memory for every time step, 
        ``default = False``"""
        self.gauge_data = []
        r"""(list) - List of numpy.ndarray objects. Each element of the list
        stores the values of the corresponding gauge if ``keep_gauges`` is set
        to ``True``"""

        self._init_q_da(num_eqn)
        if num_aux>0: self._init_aux_da(num_aux)

    def _init_aux_da(self,num_aux,num_ghost=0):
        r"""
        Initializes PETSc DA and global & local Vectors for handling the
        auxiliary array, aux. 
        
        Initializes aux_da, gauxVec and _aux_local_vector.
        """
        self.aux_da = self._create_DA(num_aux,num_ghost)
        self.gauxVec = self.aux_da.createGlobalVector()
        self._aux_local_vector = self.aux_da.createLocalVector()
 
    def _init_q_da(self,num_eqn,num_ghost=0):
        r"""
        Initializes PETSc DA and Vecs for handling the solution, q. 
        
        Initializes q_da, gqVec and _q_local_vector.
        """
        self.q_da = self._create_DA(num_eqn,num_ghost)
        self.gqVec = self.q_da.createGlobalVector()
        self._q_local_vector = self.q_da.createLocalVector()

    def _create_DA(self,dof,num_ghost=0):
        r"""Returns a PETSc DA and associated global Vec.
        Note that no local vector is returned.
        """
        from petsc4py import PETSc

        #Due to the way PETSc works, we just make the patch always periodic,
        #regardless of the boundary conditions actually selected.
        #This works because in solver.qbc() we first call globalToLocal()
        #and then impose the real boundary conditions (if non-periodic).

        if hasattr(PETSc.DA, 'PeriodicType'):
            if self.num_dim == 1:
                periodic_type = PETSc.DA.PeriodicType.X
            elif self.num_dim == 2:
                periodic_type = PETSc.DA.PeriodicType.XY
            elif self.num_dim == 3:
                periodic_type = PETSc.DA.PeriodicType.XYZ
            else:
                raise Exception("Invalid number of dimensions")

            DA = PETSc.DA().create(dim=self.num_dim,
                                          dof=dof,
                                          sizes=self.patch.num_cells_global,
                                          periodic_type = periodic_type,
                                          stencil_width=num_ghost,
                                          comm=PETSc.COMM_WORLD)
        else:
            DA = PETSc.DA().create(dim=self.num_dim,
                                          dof=dof,
                                          sizes=self.patch.num_cells_global,
                                          boundary_type = PETSc.DA.BoundaryType.PERIODIC,
                                          stencil_width=num_ghost,
                                          comm=PETSc.COMM_WORLD)

        return DA

    def get_qbc_from_q(self,num_ghost,qbc):
        """
        Returns q with ghost cells attached, by accessing the local vector.
        """
        shape = [n + 2*num_ghost for n in self.grid.num_cells]
        
        self.q_da.globalToLocal(self.gqVec, self._q_local_vector)
        shape.insert(0,self.num_eqn)
        return self._q_local_vector.getArray().reshape(shape, order = 'F')
            
    def get_auxbc_from_aux(self,num_ghost,auxbc):
        """
        Returns aux with ghost cells attached, by accessing the local vector.
        """
        shape = [n + 2*num_ghost for n in self.grid.num_cells]
        
        self.aux_da.globalToLocal(self.gauxVec, self._aux_local_vector)
        shape.insert(0,self.num_aux)
        return self._aux_local_vector.getArray().reshape(shape, order = 'F')

    def set_num_ghost(self,num_ghost):
        r"""
        This is a hack to deal with the fact that petsc4py
        doesn't allow us to change the stencil_width (num_ghost).

        Instead, we initially create DAs with stencil_width=0.
        Then, in solver.setup(), we call this function to replace
        those DAs with new ones that have the right stencil width.

        This could be made more efficient using some PETSc calls,
        but it only happens once so it seems not to be worth it.
        """
        q0 = self.q.copy()
        self._init_q_da(self.num_eqn,num_ghost)
        self.q = q0

        if self.aux is not None:
            aux0 = self.aux.copy()
            self._init_aux_da(self.num_aux,num_ghost)
            self.aux = aux0

    def sum_F(self,i):
        return self.gFVec.strideNorm(i,0)

    def get_q_global(self):
        r"""
        Returns a copy of the global q array on process 0, otherwise returns None
        """
        from petsc4py import PETSc
        q_natural = self.q_da.createNaturalVec()
        self.q_da.globalToNatural(self.gqVec, q_natural)
        scatter, q0Vec = PETSc.Scatter.toZero(q_natural)
        scatter.scatter(q_natural, q0Vec, False, PETSc.Scatter.Mode.FORWARD)
        rank = PETSc.COMM_WORLD.getRank()
        if rank == 0:
            shape = self.patch.num_cells_global
            shape.insert(0,self.num_eqn)
            q0=q0Vec.getArray().reshape(shape, order = 'F').copy()
        else:
            q0=None
        
        scatter.destroy()
        q0Vec.destroy()

        return q0

    def get_aux_global(self):
        r"""
        Returns a copy of the global aux array on process 0, otherwise returns None
        """
        from petsc4py import PETSc
        aux_natural = self.aux_da.createNaturalVec()
        self.aux_da.globalToNatural(self.gauxVec, aux_natural)
        scatter, aux0Vec = PETSc.Scatter.toZero(aux_natural)
        scatter.scatter(aux_natural, aux0Vec, False, PETSc.Scatter.Mode.FORWARD)
        rank = PETSc.COMM_WORLD.getRank()
        if rank == 0:
            shape = self.patch.num_cells_global
            shape.insert(0,self.num_aux)
            aux0=aux0Vec.getArray().reshape(shape, order = 'F').copy()
        else:
            aux0=None
        
        scatter.destroy()
        aux0Vec.destroy()

        return aux0

    def __deepcopy__(self,memo={}):
        r"""
        Calls the pyclaw deepcopy function, but also copies the number of ghost cells   
        """
        result = super(State,self).__deepcopy__(memo)
        result.set_num_ghost(self.q_da.stencil_width) 
        
        return result



r"""
Module specifying the interface to every solver in PyClaw.
"""
import logging
import numpy as np

class CFLError(Exception):
    """Error raised when cfl_max is exceeded.  Is this a
       reasonable mechanism for handling that?"""
    def __init__(self,msg):
        super(CFLError,self).__init__(msg)

class BC():
    """Enumeration of boundary condition names."""
    # This could instead just be implemented as a static dictionary.
    custom     = 0
    extrap    = 1
    periodic   = 2
    wall = 3

# =================== Dummy routines =============
def default_compute_gauge_values(q,aux):
    r"""By default, record values of q at gauges.
    """
    return q

def before_step(solver,solution):
    r"""
    Dummy routine called before each step

    Replace this routine if you want to do something before each time step.
    """
    pass

class Solver(object):
    r"""
    Pyclaw solver superclass.

    The pyclaw.Solver.solver class is an abstract class that should
    not be instantiated; rather, all Solver classes should inherit from it.

    Solver initialization takes one argument -- a Riemann solver:

        >>> from clawpack import pyclaw, riemann
        >>> solver = pyclaw.ClawSolver2D(riemann.euler_4wave_2D)

    after which solver options may be set.
    It is necessary to set the boundary conditions (for q, and
    for aux if an aux array is used):

        >>> solver.bc_lower[0] = pyclaw.BC.extrap
        >>> solver.bc_upper[0] = pyclaw.BC.wall

    Many other options may be set
    for specific solvers; for instance the limiter to be used, whether to
    use a dimensionally-split algorithm, and so forth.

    Usually the solver is attached to a controller before being used::

        >>> claw = pyclaw.Controller()
        >>> claw.solver = solver

    .. attribute:: dt

        Current time step, ``default = 0.1``

    .. attribute:: cfl

        Current Courant-Freidrichs-Lewy number, ``default = 1.0``

    .. attribute:: status

        Dictionary of status values for the solver with the following keys:
         - ``cflmax`` = Maximum CFL number
         - ``dtmin`` = Minimum time step taken
         - ``dtmax`` = Maximum time step taken
         - ``numsteps`` = Total number of time steps that have been taken

        solver.status is returned by solver.evolve_to_time.

    .. attribute:: before_step

        Function called before each time step is taken.
        The required signature for this function is:

        def before_step(solver,solution)

    .. attribute:: dt_variable

        Whether to allow the time step to vary, ``default = True``.
        If false, the initial time step size is used for all steps.

    .. attribute:: max_steps

        The maximum number of time steps allowed to reach the end time
        requested, ``default = 10000``.  If exceeded, an exception is
        raised.

    .. attribute:: logger

        Default logger for all solvers.  Records information about the run
        and debugging messages (if requested).

    .. attribute:: bc_lower

        (list of ints) Lower boundary condition types, listed in the
        same order as the Dimensions of the Patch.  See Solver.BC for
        an enumeration.

    .. attribute:: bc_upper

        (list of ints) Upper boundary condition types, listed in the
        same order as the Dimensions of the Patch.  See Solver.BC for
        an enumeration.

    .. attribute:: user_bc_lower

        (func) User defined lower boundary condition.
        Fills the values of qbc with the correct boundary values.
        The appropriate signature is:

        def user_bc_lower(patch,dim,t,qbc,num_ghost):

    .. attribute:: user_bc_upper

        (func) User defined upper boundary condition.
        Fills the values of qbc with the correct boundary values.
        The appropriate signature is:

        def user_bc_upper(patch,dim,t,qbc,num_ghost):


    :Initialization:

    Output:
     - (:class:`Solver`) - Initialized Solver object
    """

    def __setattr__(self, key, value):
        if not hasattr(self, '_isinitialized'):
            self.__dict__['_isinitialized'] = False
        if self._isinitialized and not hasattr(self, key):
            raise TypeError("%s has no attribute %s" % (self.__class__,key))
        object.__setattr__(self,key,value)

    @property
    def all_bcs(self):
        return self.bc_lower, self.bc_upper
    @all_bcs.setter
    def all_bcs(self,all_bcs):
        for i in range(self.num_dim):
            self.bc_lower[i] = all_bcs
            self.bc_upper[i] = all_bcs

    #  ======================================================================
    #   Initialization routines
    #  ======================================================================
    def __init__(self,riemann_solver=None,claw_package=None):
        r"""
        Initialize a Solver object

        See :class:`Solver` for full documentation
        """
        # Setup solve logger
        self.logger = logging.getLogger('pyclaw.solver')

        self.dt_initial = 0.1
        self.dt_max = 1e99
        self.max_steps = 10000
        self.dt_variable = True
        self.num_waves = None  # Must be set later to agree with Riemann solver
        self.qbc = None
        self.auxbc = None
        self.rp = None
        self.fmod = None
        self._is_set_up = False
        self._use_old_bc_sig = False
        self.accept_step = True
        self.before_step = None

        # select package to build solver objects from, by default this will be
        # the package that contains the module implementing the derived class
        # for example, if ClawSolver1D is implemented in 'clawpack.petclaw.solver', then
        # the computed claw_package will be 'clawpack.petclaw'

        import sys
        if claw_package is not None and claw_package in sys.modules:
            self.claw_package = sys.modules[claw_package]
        else:
            def get_clawpack_dot_xxx(modname): return modname.rpartition('.')[0].rpartition('.')[0]
            claw_package_name = get_clawpack_dot_xxx(self.__module__)
            if claw_package_name in sys.modules:
                self.claw_package = sys.modules[claw_package_name]
            else:
                raise NotImplementedError("Unable to determine solver package, please provide one")

        # Initialize time stepper values
        self.dt = self.dt_initial
        self.cfl = self.claw_package.CFL(self.cfl_desired)

        # Status Dictionary
        self.status = {'cflmax': -np.inf,
                       'dtmin': np.inf,
                       'dtmax': -np.inf,
                       'numsteps': 0 }

        # No default BCs; user must set them
        self.bc_lower =    [None]*self.num_dim
        self.bc_upper =    [None]*self.num_dim
        self.aux_bc_lower = [None]*self.num_dim
        self.aux_bc_upper = [None]*self.num_dim

        self.user_bc_lower = None
        self.user_bc_upper = None

        self.user_aux_bc_lower = None
        self.user_aux_bc_upper = None

        self.num_eqn   = None
        self.num_waves = None

        self.compute_gauge_values = default_compute_gauge_values
        r"""(function) - Function that computes quantities to be recorded at gauges"""

        self.qbc          = None
        r""" Array to hold ghost cell values.  This is the one that gets passed
        to the Fortran code.  """

        if riemann_solver is not None:
            self.rp = riemann_solver
            rp_name = riemann_solver.__name__.split('.')[-1]
            from clawpack import riemann
            if "ptwise" in rp_name:
                rp_name = rp_name.replace("_ptwise", "")
            self.num_eqn   = riemann.static.num_eqn.get(rp_name,None)
            self.num_waves = riemann.static.num_waves.get(rp_name,None)

        self._isinitialized = True

        super(Solver,self).__init__()

    # ========================================================================
    #  Solver setup and validation routines
    # ========================================================================
    def is_valid(self):
        r"""
        Checks that all required solver attributes are set.

        Checks to make sure that all the required attributes for the solver
        have been set correctly.  All required attributes that need to be set
        are contained in the attributes list of the class.

        Will post debug level logging message of which required attributes
        have not been set.

        :Output:
         - *valid* - (bool) True if the solver is valid, False otherwise

        """
        valid = True
        reason = None
        if any([bcmeth == BC.custom for bcmeth in self.bc_lower]):
            if self.user_bc_lower is None:
                valid = False
                reason = 'Lower custom BC function has not been set.'
        if any([bcmeth == BC.custom for bcmeth in self.bc_upper]):
            if self.user_bc_upper is None:
                valid = False
                reason = 'Upper custom BC function has not been set.'
        if self.num_waves is None:
            valid = False
            reason = 'solver.num_waves has not been set.'
        if self.num_eqn is None:
            valid = False
            reason = 'solver.num_eqn has not been set.'
        if (None in self.bc_lower) or (None in self.bc_upper):
            valid = False
            reason = 'One of the boundary conditions has not been set.'

        if reason is not None:
            self.logger.debug(reason)
        return valid, reason

    def setup(self,solution):
        r"""
        Stub for solver setup routines.

        This function is called before a set of time steps are taken in order
        to reach tend.  A subclass should extend or override it if it needs to
        perform some setup based on attributes that would be set after the
        initialization routine.  Typically this is initialization that
        requires knowledge of the solution object.
        """

        self._is_set_up = True

    def __del__(self):
        r"""
        Stub for solver teardown routines.

        This function is called at the end of a simulation.
        A subclass should override it only if it needs to
        perform some cleanup, such as deallocating arrays in a Fortran module.
        """
        self._is_set_up = False

    def __str__(self):
        output = "Solver Status:\n"
        for (k,v) in self.status.items():
            output = "\n".join((output,"%s = %s" % (k.rjust(25),v)))
        return output

    # ========================================================================
    #  Boundary Conditions
    # ========================================================================
    def _allocate_bc_arrays(self,state):
        r"""
        Create numpy arrays for q and aux with ghost cells attached.
        These arrays are referred to throughout the code as qbc and auxbc.

        This is typically called by solver.setup().
        """
        import inspect
        for fun in (self.user_bc_lower,self.user_bc_upper,self.user_aux_bc_lower,self.user_aux_bc_upper):
            if fun is not None:
                args = inspect.getfullargspec(fun)[0]
                if len(args) == 5:
                    self.logger.warn("""The custom boundary condition
                                        function signature has been changed.
                                        The previous signature will not be
                                        supported in Clawpack 6.0.  Please see
                                        http://www.clawpack.org/pyclaw/solvers.html#change-to-custom-bc-function-signatures
                                        for more information.""")
                    self._use_old_bc_sig = True

        qbc_dim = [n+2*self.num_ghost for n in state.grid.num_cells]
        qbc_dim.insert(0,state.num_eqn)
        self.qbc = np.zeros(qbc_dim,order='F')

        auxbc_dim = [n+2*self.num_ghost for n in state.grid.num_cells]
        auxbc_dim.insert(0,state.num_aux)
        self.auxbc = np.empty(auxbc_dim,order='F')

        self._apply_bcs(state)

    def _apply_bcs(self, state):
        r"""
        Apply boundary conditions to both q and aux arrays.

        In the case of a user-defined boundary condition, both arrays
        qbc and auxbc are passed to the user function.  Typically the
        function would only modify one or the other of them, though this
        is not enforced.

        If the user function only accepts one array argument, we warn
        that this interface has been deprecated.  In Clawpack 6, we will
        drop backward compatibility.

        For parallel runs, we check whether we're actually on a domain
        boundary.  If we are just at an inter-patch boundary, nothing needs to
        be done here.
        """

        import numpy as np

        self.qbc = state.get_qbc_from_q(self.num_ghost, self.qbc)
        if state.num_aux > 0:
            self.auxbc = state.get_auxbc_from_aux(self.num_ghost, self.auxbc)

        grid = state.grid

        for (idim, dim) in enumerate(grid.dimensions):
            # Check if we are on a true boundary
            if state.grid.on_lower_boundary[idim]:

                bcs = []
                if state.num_aux > 0:
                    bcs.append({'array'  : self.auxbc,
                                'type'   : self.aux_bc_lower,
                                'custom_fun' : self.user_aux_bc_lower,
                                'variable' : 'aux'})
                bcs.append({'array'  : self.qbc,
                            'type'   : self.bc_lower,
                            'custom_fun' : self.user_bc_lower,
                            'variable' : 'q'})
                for (i, bc) in enumerate(bcs):

                    if bc['type'][idim] == BC.custom:
                        if not self._use_old_bc_sig:
                            bc['custom_fun'](state, dim, state.t, self.qbc,
                                             self.auxbc, self.num_ghost)
                        else:
                            bc['custom_fun'](state, dim, state.t, bc['array'], self.num_ghost)

                    elif bc['type'][idim] == BC.periodic \
                            and not state.grid.on_upper_boundary[idim]:
                        pass  # In a parallel run, # PETSc handles periodic BCs.

                    else:
                        self._bc_lower(bc['type'][idim], state, dim, state.t,
                                        np.rollaxis(bc['array'], idim+1, 1), idim,
                                        bc['variable'])

            if state.grid.on_upper_boundary[idim]:

                bcs = []
                if state.num_aux > 0:
                    bcs.append({'array'  : self.auxbc,
                                'type'   : self.aux_bc_upper,
                                'custom_fun' : self.user_aux_bc_upper,
                                'variable' : 'aux'})
                bcs.append({'array'  : self.qbc,
                            'type'   : self.bc_upper,
                            'custom_fun' : self.user_bc_upper,
                            'variable' : 'q'})
                for (i, bc) in enumerate(bcs):

                    if bc['type'][idim] == BC.custom:
                        if not self._use_old_bc_sig:
                            bc['custom_fun'](state, dim, state.t, self.qbc,
                                             self.auxbc, self.num_ghost)
                        else:
                            bc['custom_fun'](state, dim, state.t, bc['array'], self.num_ghost)

                    elif bc['type'][idim] == BC.periodic \
                            and not state.grid.on_lower_boundary[idim]:
                        pass  # In a parallel run, # PETSc handles periodic BCs.

                    else:
                        self._bc_upper(bc['type'][idim], state, dim, state.t,
                                        np.rollaxis(bc['array'], idim+1, 1), idim,
                                        bc['variable'])

    def _bc_lower(self, bc_type, state, dim, t, array, idim, name):
        r"""
        Apply lower boundary conditions to array.

        Sets the lower coordinate's ghost cells of *array* depending on what
        :attr:`bc_lower` is.  If :attr:`bc_lower` = 0 then the user
        boundary condition specified by :attr:`user_bc_lower` is used.  Note
        that in this case the function :attr:`user_bc_lower` belongs only to
        this dimension but :attr:`user_bc_lower` could set all user boundary
        conditions at once with the appropriate calling sequence.

        :Input:
         - *patch* - (:class:`Patch`) Patch that the dimension belongs to.

        :Input/Output:
         - *array* - (ndarray(...,num_eqn)) Array with added ghost cells which
           will be set in this routines.
        """

        if bc_type == BC.extrap:
            for i in range(self.num_ghost):
                array[:,i,...] = array[:,self.num_ghost,...]
        elif bc_type == BC.periodic:
            # This process owns the whole patch
            array[:,:self.num_ghost,...] = array[:,-2*self.num_ghost:-self.num_ghost,...]
        elif bc_type == BC.wall:
            if name == 'q':
                for i in range(self.num_ghost):
                    array[:,i,...] = array[:,2*self.num_ghost-1-i,...]
                    # Negate normal velocity
                    array[self.reflect_index[idim],i,...] = \
                        -array[self.reflect_index[idim],2*self.num_ghost-1-i,...]
            else:
                for i in range(self.num_ghost):
                    array[:,i,...] = array[:,2*self.num_ghost-1-i,...]
        else:
            if bc_type is None:
                raise Exception('Lower boundary condition not specified for either q or aux.')
            else:
                raise NotImplementedError("Boundary condition %s not implemented" % bc_type)

    def _bc_upper(self, bc_type, state, dim, t, array, idim, name):
        r"""
        Apply upper boundary conditions to array

        Sets the upper coordinate's ghost cells of *array* depending on what
        :attr:`bc_upper` is.  If :attr:`bc_upper` = 0 then the user
        boundary condition specified by :attr:`user_bc_upper` is used.  Note
        that in this case the function :attr:`user_bc_upper` belongs only to
        this dimension but :attr:`user_bc_upper` could set all user boundary
        conditions at once with the appropriate calling sequence.

        :Input:
         - *patch* - (:class:`Patch`) Patch that the dimension belongs to

        :Input/Output:
         - *array* - (ndarray(...,num_eqn)) Array with added ghost cells which will
           be set in this routines
        """

        if bc_type == BC.extrap:
            for i in range(self.num_ghost):
                array[:,-i-1,...] = array[:,-self.num_ghost-1,...]
        elif bc_type == BC.periodic:
            # This process owns the whole patch
            array[:,-self.num_ghost:,...] = array[:,self.num_ghost:2*self.num_ghost,...]
        elif bc_type == BC.wall:
            if name == 'q':
                for i in range(self.num_ghost):
                    array[:,-i-1,...] = array[:,-2*self.num_ghost+i,...]
                    #  Negate normal velocity
                    array[self.reflect_index[idim],-i-1,...] = \
                        -array[self.reflect_index[idim],-2*self.num_ghost+i,...]
            else:
                for i in range(self.num_ghost):
                    array[:,-i-1,...] = array[:,-2*self.num_ghost+i,...]
        else:
            if bc_type is None:
                raise Exception('Upper boundary condition not specified for either q or aux.')
            else:
                raise NotImplementedError("Boundary condition %s not implemented" % bc_type)

    # ========================================================================
    #  Evolution routines
    # ========================================================================
    def accept_reject_step(self,state):
        cfl = self.cfl.get_cached_max()
        if cfl > self.cfl_max:
            return False
        else:
            return True

    def get_dt_new(self):
        cfl = self.cfl.get_cached_max()
        self.dt = min(self.dt_max,self.dt * self.cfl_desired / cfl)

    def get_dt(self,t,tstart,tend,take_one_step):
        cfl = self.cfl.get_cached_max()
        if self.dt_variable and self.dt_old is not None:
            if cfl > 0.0:
                self.get_dt_new()
                self.status['dtmin'] = min(self.dt, self.status['dtmin'])
                self.status['dtmax'] = max(self.dt, self.status['dtmax'])
            else:
                self.dt = self.dt_max
        else:
            self.dt_old = self.dt

        # Adjust dt so that we hit tend exactly if we are near tend
        if not take_one_step:
            if t + self.dt > tend and tstart < tend:
                self.dt = tend - t
            if tend - t - self.dt < 1.e-14*t:
                self.dt = tend - t

    def evolve_to_time(self, solution, tend=None):
        r"""
        Evolve solution from solution.t to tend.  If tend is not specified,
        take a single step.

        This method contains the machinery to evolve the solution object in
        ``solution`` to the requested end time tend if given, or one
        step if not.

        :Input:
         - *solution* - (:class:`Solution`) Solution to be evolved
         - *tend* - (float) The end time to evolve to, if not provided then
           the method will take a single time step.

        :Output:
         - (dict) - Returns the status dictionary of the solver
        """

        if not self._is_set_up:
            self.setup(solution)

        if tend is None:
            take_one_step = True
        else:
            take_one_step = False

        # Parameters for time-stepping
        tstart = solution.t

        num_steps = 0

        # Setup for the run
        if not self.dt_variable:
            if take_one_step:
                self.max_steps = 1
            else:
                self.max_steps = int((tend - tstart + 1e-10) / self.dt)
                if abs(self.max_steps*self.dt - (tend - tstart)) >      \
                   1e-5 * (tend - tstart):
                    raise Exception('dt does not divide (tend-tstart) and dt '
                                    'is fixed!')
        if self.dt_variable == 1 and self.cfl_desired > self.cfl_max:
            raise Exception('Variable time-stepping and desired CFL > maximum '
                            'CFL')
        if not take_one_step:
            if tend <= tstart:
                self.logger.info("Already at or beyond end time: no evolution ",
                                 "required.")
                self.max_steps = 0

        # Main time-stepping loop
        for n in range(self.max_steps):

            state = solution.state

            # Keep a backup in case we need to retake a time step
            if self.dt_variable:
                q_backup = state.q.copy('F')
                told = solution.t

            if self.before_step is not None:
                self.before_step(self, solution.states[0])

            # Note that the solver may alter dt during the step() routine
            self.step(solution, take_one_step, tstart, tend)

            # Check to make sure that the Courant number was not too large
            cfl = self.cfl.get_cached_max()
            self.accept_step = self.accept_reject_step(state)
            if self.accept_step:
                # Accept this step
                self.status['cflmax'] = max(cfl, self.status['cflmax'])
                if self.dt_variable:
                    solution.t += self.dt
                else:
                    # Avoid roundoff error if dt_variable==False:
                    solution.t = tstart+(n+1)*self.dt

                # Verbose messaging
                self.logger.debug("Step %i  CFL = %f   dt = %f   t = %f"
                                  % (n,cfl,self.dt,solution.t))

                self.write_gauge_values(solution)
                # Increment number of time steps completed
                num_steps += 1
                self.status['numsteps'] += 1

            else:
                # Reject this step
                self.logger.debug("Rejecting time step, CFL number too large")
                if self.dt_variable:
                    state.q = q_backup
                    solution.t = told
                else:
                    # Give up, we cannot adapt, abort
                    self.status['cflmax'] = \
                        max(cfl, self.status['cflmax'])
                    raise Exception('CFL too large, giving up!')

            # See if we are finished yet
            if take_one_step:
                break
            elif solution.t >= tend:
                break

        # End of main time-stepping loop -------------------------------------

        if not take_one_step:
            if self.dt_variable and solution.t < tend \
                    and num_steps == self.max_steps:
                raise Exception("Maximum number of timesteps have been taken")

        return self.status

    def step(self,solution):
        r"""
        Take one step

        This method is only a stub and should be overridden by all solvers who
        would like to use the default time-stepping in evolve_to_time.
        """
        raise NotImplementedError("No stepping routine has been defined!")

    # ========================================================================
    #  Gauges
    # ========================================================================
    def write_gauge_values(self,solution):
        r"""Write solution (or derived quantity) values at each gauge coordinate
            to file.
        """
        import numpy as np
        if solution.num_aux == 0:
            aux = None
        for i,gauge in enumerate(solution.state.grid.gauges):
            if self.num_dim == 1:
                ix=gauge[0]
                if solution.num_aux > 0:
                    aux = solution.state.aux[:,ix]
                q=solution.state.q[:,ix]
            elif self.num_dim == 2:
                ix, iy = gauge
                if solution.num_aux > 0:
                    aux = solution.state.aux[:,ix,iy]
                q=solution.state.q[:,ix,iy]
            p=self.compute_gauge_values(q,aux)
            if not hasattr(p,'__iter__'):
                p = [p]
            t=solution.t
            if solution.state.keep_gauges:
                gauge_data = solution.state.gauge_data
                if len(gauge_data) == len(solution.state.grid.gauges):
                    gauge_data[i]=np.vstack((gauge_data[i],np.append(t,p)))
                else:
                    gauge_data.append(np.append(t,p))

            try:
                solution.state.grid.gauge_files[i].write(str(t)+' '+' '.join(str(j)
                                                         for j in p)+'\n')
            except IOError:
                raise Exception("Gauge files are not set up correctly. You should call \
                       \nthe method `setup_gauge_files` of the Grid class object \
                       \nbefore any call for `write_gauge_values` from the Solver class.")


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# encoding: utf-8
"""
Module to set up run time parameters for Clawpack.

The values set in the function setrun are then written out to data files
that will be read in by the Fortran code.

"""

from __future__ import absolute_import
from __future__ import print_function
import os
import datetime

import numpy as np

#                           days   s/hour    hours/day            
days2seconds = lambda days: days * 60.0**2 * 24.0
seconds2days = lambda seconds: seconds / (60.0**2 * 24.0)

#------------------------------
def setrun(claw_pkg='geoclaw'):
#------------------------------

    """
    Define the parameters used for running Clawpack.

    INPUT:
        claw_pkg expected to be "geoclaw" for this setrun.

    OUTPUT:
        rundata - object of class ClawRunData

    """

    from clawpack.clawutil import data

    assert claw_pkg.lower() == 'geoclaw',  "Expected claw_pkg = 'geoclaw'"

    num_dim = 2
    rundata = data.ClawRunData(claw_pkg, num_dim)

    #------------------------------------------------------------------
    # Problem-specific parameters to be written to setprob.data:
    #------------------------------------------------------------------
    
    #probdata = rundata.new_UserData(name='probdata',fname='setprob.data')

    #------------------------------------------------------------------
    # Standard Clawpack parameters to be written to claw.data:
    #   (or to amr2ez.data for AMR)
    #------------------------------------------------------------------
    clawdata = rundata.clawdata  # initialized when rundata instantiated


    # Set single grid parameters first.
    # See below for AMR parameters.


    # ---------------
    # Spatial domain:
    # ---------------

    # Number of space dimensions:
    clawdata.num_dim = num_dim

    # Lower and upper edge of computational domain:
    clawdata.lower[0] = -99.0      # west longitude
    clawdata.upper[0] = -70.0      # east longitude

    clawdata.lower[1] = 8.0       # south latitude
    clawdata.upper[1] = 32.0      # north latitude

    # Number of grid cells:
    degree_factor = 4 # (0.25º,0.25º) ~ (25237.5 m, 27693.2 m) resolution
    clawdata.num_cells[0] = int(clawdata.upper[0] - clawdata.lower[0]) * degree_factor
    clawdata.num_cells[1] = int(clawdata.upper[1] - clawdata.lower[1]) * degree_factor

    # ---------------
    # Size of system:
    # ---------------

    # Number of equations in the system:
    clawdata.num_eqn = 3

    # Number of auxiliary variables in the aux array (initialized in setaux)
    # First three are from shallow GeoClaw, fourth is friction and last 3 are
    # storm fields
    clawdata.num_aux = 3 + 1 + 3

    # Index of aux array corresponding to capacity function, if there is one:
    clawdata.capa_index = 2

    
    
    # -------------
    # Initial time:
    # -------------
    # read_atcf currently just assumes a time_offset of the first recorded time
    # so this is done manually
    clawdata.t0 = 9.5e5

    # Restart from checkpoint file of a previous run?
    # Note: If restarting, you must also change the Makefile to set:
    #    RESTART = True
    # If restarting, t0 above should be from original run, and the
    # restart_file 'fort.chkNNNNN' specified below should be in 
    # the OUTDIR indicated in Makefile.

    clawdata.restart = False               # True to restart from prior results
    clawdata.restart_file = 'fort.chk00006'  # File to use for restart data

    # -------------
    # Output times:
    #--------------

    # Specify at what times the results should be written to fort.q files.
    # Note that the time integration stops after the final output time.
    # The solution at initial time t0 is always written in addition.

    clawdata.output_style = 1

    if clawdata.output_style == 1:
        # Output nout frames at equally spaced times up to tfinal:
        clawdata.tfinal = 9.8e5
        recurrence = 2
        clawdata.num_output_times = int((clawdata.tfinal - clawdata.t0) 
                                            * recurrence / (60**2 * 24))

        clawdata.output_t0 = True  # output at initial (or restart) time?
        

    elif clawdata.output_style == 2:
        # Specify a list of output times.
        clawdata.output_times = [0.5, 1.0]

    elif clawdata.output_style == 3:
        # Output every iout timesteps with a total of ntot time steps:
        clawdata.output_step_interval = 1
        clawdata.total_steps = 1
        clawdata.output_t0 = True
        

    clawdata.output_format = 'binary'      # 'ascii' or 'netcdf' 
    clawdata.output_q_components = 'all'   # could be list such as [True,True]
    clawdata.output_aux_components = 'all'
    clawdata.output_aux_onlyonce = False    # output aux arrays only at t0



    # ---------------------------------------------------
    # Verbosity of messages to screen during integration:
    # ---------------------------------------------------

    # The current t, dt, and cfl will be printed every time step
    # at AMR levels <= verbosity.  Set verbosity = 0 for no printing.
    #   (E.g. verbosity == 2 means print only on levels 1 and 2.)
    clawdata.verbosity = 1



    # --------------
    # Time stepping:
    # --------------

    # if dt_variable==1: variable time steps used based on cfl_desired,
    # if dt_variable==0: fixed time steps dt = dt_initial will always be used.
    clawdata.dt_variable = True

    # Initial time step for variable dt.
    # If dt_variable==0 then dt=dt_initial for all steps:
    clawdata.dt_initial = 0.016

    # Max time step to be allowed if variable dt used:
    clawdata.dt_max = 1e+99

    # Desired Courant number if variable dt used, and max to allow without
    # retaking step with a smaller dt:
    clawdata.cfl_desired = 0.75
    clawdata.cfl_max = 1.0
    # clawdata.cfl_desired = 0.25
    # clawdata.cfl_max = 0.5

    # Maximum number of time steps to allow between output times:
    clawdata.steps_max = 5000




    # ------------------
    # Method to be used:
    # ------------------

    # Order of accuracy:  1 => Godunov,  2 => Lax-Wendroff plus limiters
    clawdata.order = 1
    
    # Use dimensional splitting? (not yet available for AMR)
    clawdata.dimensional_split = 'unsplit'
    
    # For unsplit method, transverse_waves can be 
    #  0 or 'none'      ==> donor cell (only normal solver used)
    #  1 or 'increment' ==> corner transport of waves
    #  2 or 'all'       ==> corner transport of 2nd order corrections too
    clawdata.transverse_waves = 1

    # Number of waves in the Riemann solution:
    clawdata.num_waves = 3
    
    # List of limiters to use for each wave family:  
    # Required:  len(limiter) == num_waves
    # Some options:
    #   0 or 'none'     ==> no limiter (Lax-Wendroff)
    #   1 or 'minmod'   ==> minmod
    #   2 or 'superbee' ==> superbee
    #   3 or 'mc'       ==> MC limiter
    #   4 or 'vanleer'  ==> van Leer
    clawdata.limiter = ['mc', 'mc', 'mc']

    clawdata.use_fwaves = True    # True ==> use f-wave version of algorithms
    
    # Source terms splitting:
    #   src_split == 0 or 'none'    ==> no source term (src routine never called)
    #   src_split == 1 or 'godunov' ==> Godunov (1st order) splitting used, 
    #   src_split == 2 or 'strang'  ==> Strang (2nd order) splitting used,  not recommended.
    clawdata.source_split = 'godunov'
    # clawdata.source_split = 'strang'


    # --------------------
    # Boundary conditions:
    # --------------------

    # Number of ghost cells (usually 2)
    clawdata.num_ghost = 2

    # Choice of BCs at xlower and xupper:
    #   0 => user specified (must modify bcN.f to use this option)
    #   1 => extrapolation (non-reflecting outflow)
    #   2 => periodic (must specify this at both boundaries)
    #   3 => solid wall for systems where q(2) is normal velocity

    clawdata.bc_lower[0] = 'extrap'
    clawdata.bc_upper[0] = 'extrap'

    clawdata.bc_lower[1] = 'extrap'
    clawdata.bc_upper[1] = 'extrap'

    # Specify when checkpoint files should be created that can be
    # used to restart a computation.

    clawdata.checkpt_style = 0

    if clawdata.checkpt_style == 0:
        # Do not checkpoint at all
        pass

    elif clawdata.checkpt_style == 1:
        # Checkpoint only at tfinal.
        pass

    elif clawdata.checkpt_style == 2:
        # Specify a list of checkpoint times.  
        clawdata.checkpt_times = [0.1,0.15]

    elif clawdata.checkpt_style == 3:
        # Checkpoint every checkpt_interval timesteps (on Level 1)
        # and at the final time.
        clawdata.checkpt_interval = 5


    # ---------------
    # AMR parameters:
    # ---------------
    amrdata = rundata.amrdata

    # max number of refinement levels:
    amrdata.amr_levels_max = 2

    # List of refinement ratios at each level (length at least mxnest-1)
    amrdata.refinement_ratios_x = [2,2,2,6,16]
    amrdata.refinement_ratios_y = [2,2,2,6,16]
    amrdata.refinement_ratios_t = [2,2,2,6,16]


    # Specify type of each aux variable in amrdata.auxtype.
    # This must be a list of length maux, each element of which is one of:
    #   'center',  'capacity', 'xleft', or 'yleft'  (see documentation).

    amrdata.aux_type = ['center','capacity','yleft','center','center','center',
                         'center', 'center', 'center']


    # Flag using refinement routine flag2refine rather than richardson error
    amrdata.flag_richardson = False    # use Richardson?
    amrdata.flag2refine = True

    # steps to take on each level L between regriddings of level L+1:
    amrdata.regrid_interval = 3

    # width of buffer zone around flagged points:
    # (typically the same as regrid_interval so waves don't escape):
    amrdata.regrid_buffer_width  = 2

    # clustering alg. cutoff for (# flagged pts) / (total # of cells refined)
    # (closer to 1.0 => more small grids may be needed to cover flagged cells)
    amrdata.clustering_cutoff = 0.700000

    # print info about each regridding up to this level:
    amrdata.verbosity_regrid = 0  


    #  ----- For developers ----- 
    # Toggle debugging print statements:
    amrdata.dprint = False      # print domain flags
    amrdata.eprint = False      # print err est flags
    amrdata.edebug = False      # even more err est flags
    amrdata.gprint = False      # grid bisection/clustering
    amrdata.nprint = False      # proper nesting output
    amrdata.pprint = False      # proj. of tagged points
    amrdata.rprint = False      # print regridding summary
    amrdata.sprint = False      # space/memory output
    amrdata.tprint = False      # time step reporting each level
    amrdata.uprint = False      # update/upbnd reporting
    
    # More AMR parameters can be set -- see the defaults in pyclaw/data.py

    # == setregions.data values ==
    # regions = rundata.regiondata.regions
    # to specify regions of refinement append lines of the form
    #  [minlevel,maxlevel,t1,t2,x1,x2,y1,y2]

    # Gauge for testing
    rundata.gaugedata.gauges.append([1, -90., 25., 
                                  rundata.clawdata.t0, rundata.clawdata.tfinal])

    #------------------------------------------------------------------
    # GeoClaw specific parameters:
    #------------------------------------------------------------------
    rundata = setgeo(rundata)
    
    #------------------------------------------------------------------
    # storm surge specific parameters:
    #------------------------------------------------------------------
    rundata = set_storm(rundata)
    
    return rundata
    # end of function setrun
    # ----------------------


#-------------------
def setgeo(rundata):
#-------------------
    """
    Set GeoClaw specific runtime parameters.
    For documentation see ....
    """

    try:
        geo_data = rundata.geo_data
    except:
        print("*** Error, this rundata has no geo_data attribute")
        raise AttributeError("Missing geo_data attribute")
       
    # == Physics ==
    geo_data.gravity = 9.81
    geo_data.coordinate_system = 2
    geo_data.earth_radius = 6367.5e3
    geo_data.rho = 1025.0
    geo_data.rho_air = 1.15
    geo_data.ambient_pressure = 101.3e3 # Nominal atmos pressure

    # == Forcing Options
    geo_data.coriolis_forcing = True
    geo_data.friction_forcing = True
    geo_data.manning_coefficient = 0.025 # Overridden below
    geo_data.friction_depth = 1e10

    # == Algorithm and Initial Conditions ==
    geo_data.sea_level = 0.28  # Due to seasonal swelling of gulf
    geo_data.dry_tolerance = 1.e-2

    # Refinement Criteria
    refine_data = rundata.refinement_data
    refine_data.wave_tolerance = 1.0
    refine_data.speed_tolerance = [1.0,2.0,3.0,4.0]
    refine_data.deep_depth = 300.0
    refine_data.max_level_deep = 4
    refine_data.variable_dt_refinement_ratios = True

    # == settopo.data values ==
    topo_data = rundata.topo_data
    topo_data.topofiles = []
    # for topography, append lines of the form
    #   [topotype, minlevel, maxlevel, t1, t2, fname]
    # See regions for control over these regions, need better bathy data for the
    # smaller domains
    topo_data.topofiles.append([3, 1, 5, rundata.clawdata.t0, rundata.clawdata.tfinal, 
                              'gulf_caribbean.tt3'])
    # == setdtopo.data values ==
    dtopo_data = rundata.dtopo_data
    dtopo_data.dtopofiles = []
    # for moving topography, append lines of the form :   (<= 1 allowed for now!)
    #   [topotype, minlevel,maxlevel,fname]

    # == setqinit.data values ==
    rundata.qinit_data.qinit_type = 0
    rundata.qinit_data.qinitfiles = []
    # for qinit perturbations, append lines of the form: (<= 1 allowed for now!)
    #   [minlev, maxlev, fname]

    # == setfixedgrids.data values ==
    rundata.fixed_grid_data.fixedgrids = []
    # for fixed grids append lines of the form
    # [t1,t2,noutput,x1,x2,y1,y2,xpoints,ypoints,\
    #  ioutarrivaltimes,ioutsurfacemax]

    return rundata
    # end of function setgeo
    # ----------------------


def set_storm(rundata):

    data = rundata.surge_data

    # Source term controls
    data.wind_forcing = True
    data.drag_law = 1
    data.pressure_forcing = True

    # AMR parameters
    data.wind_refine = [20.0,40.0,60.0] # m/s
    data.R_refine = [60.0e3,40e3,20e3]  # m
    
    # Storm parameters
    data.storm_type = 1 # Type of storm
    data.display_landfall_time = True

    # Storm type 2 - Idealized storm track
    data.storm_file = 'ike.storm'

    return rundata


def set_friction(rundata):

    data = rundata.frictiondata

    # Variable friction
    data.variable_friction = True

    # Region based friction
    # Entire domain
    data.friction_regions.append([rundata.clawdata.lower, 
                                  rundata.clawdata.upper,
                                  [np.infty,0.0,-np.infty],
                                  [0.030, 0.022]])

    # La-Tex Shelf
    data.friction_regions.append([(-98, 25.25), (-90, 30),
                                  [np.infty,-10.0,-200.0,-np.infty],
                                  [0.030, 0.012, 0.022]])

    return data


if __name__ == '__main__':
    # Set up run-time parameters and write all data files.
    import sys
    if len(sys.argv) == 2:
        rundata = setrun(sys.argv[1])
    else:
        rundata = setrun()

    rundata.write()



#!/usr/bin/env python
# encoding: utf-8

r"""Contains plane wave tests for the multi-layer SWE code

Can run a particular test by giving the number of the test at the comand line.

:Note:

This test suite requires the `batch` package.

"""

# ============================================================================
#  Distributed under the terms of the Berkeley Software Distribution (BSD) 
#  license
#                     http://www.opensource.org/licenses/
# ============================================================================

from __future__ import absolute_import
from __future__ import print_function
import sys
import numpy

import batch

import clawpack.geoclaw.topotools as tt

import setrun

class PlaneWaveTest(batch.Job):

    r"""Base plane wave test case.

    """

    def __init__(self, angle=0.0, bathy_angle=0.0, location=[-0.1, 0.0], 
                       bathy_location=0.15):

        super(PlaneWaveTest, self).__init__()

        self.executable = 'xgeoclaw'
        self.type = "multilayer"
        self.name = "planewave"

        # Convert angle to degrees for the label
        self.prefix = "ml_2d_ia%s_ba%s" % (int(angle * 180.0 / numpy.pi),
            int(bathy_angle * 180.0 / numpy.pi))

        # Data objects
        self.rundata = setrun.setrun()

        self.rundata.qinit_data.angle = angle
        self.rundata.qinit_data.init_location = location

        self.bathy_location = bathy_location
        self.bathy_angle = bathy_angle

        # Add gauges down perpendicular
        self.rundata.gaugedata.gauges = []
        gauge_locations = [-0.1,0.0,0.1,0.2,0.3]
        for (i,x_c) in enumerate(gauge_locations):
            # y0 = (self.run_data.clawdata.yupper - self.run_data.clawdata.ylower) / 2.0
            # x_p,y_p = transform_c2p(x_c,0.0,location[0],location[1],angle)
            x_p = x_c * numpy.cos(angle)
            y_p = x_c * numpy.sin(angle)
            # print "+=====+"
            # print x_c,0.0
            # print x_p,y_p
            if (self.rundata.clawdata.lower[0] < x_p < self.rundata.clawdata.upper[0] and
                    self.rundata.clawdata.lower[1] < y_p < self.rundata.clawdata.upper[1]):
                self.rundata.gaugedata.gauges.append([i, x_p, y_p, 0.0, 1e10])
                # print "Gauge %s: (%s,%s)" % (i,x_p,y_p)
        # print "+=====+"

        # self.setplot = lambda plotdata:setplot.setplot(
        #                                         plotdata, 
        #                                         bathy_location=bathy_location, 
        #                                         bathy_angle=bathy_angle)

    def __str__(self):
        output = super(PlaneWaveTest, self).__str__()
        output += "  Angle = %s\n" % self.rundata.qinit_data.angle
        output += "  Bathy Angle = %s\n" % self.bathy_angle
        return output

    def write_data_objects(self):
        super(PlaneWaveTest, self).write_data_objects()

        # Write out bathy file
        mx = self.rundata.clawdata.num_cells[0]
        my = self.rundata.clawdata.num_cells[1]
        xlower = self.rundata.clawdata.lower[0]
        xupper = self.rundata.clawdata.upper[0]
        ylower = self.rundata.clawdata.lower[1]
        yupper = self.rundata.clawdata.upper[1]
        dx = (xupper - xlower) / mx
        dy = (yupper - ylower) / my
        d = min(dx,dy)
        mx = int((xupper - xlower) / d) + 8
        my = int((yupper - ylower) / d) + 8
        
        xlower = xlower - d*4.0
        ylower = ylower - d*4.0
        xupper = xupper + d*4.0
        yupper = yupper + d*4.0

        step = lambda x,y: setrun.bathy_step(x, y, location=self.bathy_location,
                                            angle=self.bathy_angle)
        
        tt.topo2writer('./topo.tt2', step, xlower, xupper, ylower, yupper, 
                                     mx, my, nodata_value=-99999)

        # Write out simple bathy geometry file for communication to the plotting
        with open("./bathy_geometry.data", 'w') as bathy_geometry_file:
            bathy_geometry_file.write("%s\n%s" % (self.bathy_location,
                                                  self.bathy_angle) )


class BubbleTest(PlaneWaveTest):

    r"""Jump bathymetry with gaussian bump initial conditions."""

    def __init__(self, eigen_method=1):

        super(BubbleTest, self).__init__()

        self.type = "multilayer"
        self.name = "bubble"
        self.prefix = "ml_e%s" % eigen_method

        # Data objects
        self.rundata = setrun.setrun()

        self.rundata.clawdata.lower = [0.0, 0.0]
        self.rundata.clawdata.upper = [1.0, 1.0]
        self.rundata.clawdata.num_cells = [100, 100]
        self.rundata.multilayer_data.eigen_method = eigen_method
        self.rundata.multilayer_data.eta = [0.0, -0.5]

        self.rundata.qinit_data.qinit_type = 7
        self.rundata.qinit_data.init_location = [-0.25,0.0]
        self.rundata.qinit_data.wave_family = 0
        self.rundata.qinit_data.epsilon = 0.4
        self.rundata.qinit_data.sigma = 0.08

        self.bathy_location = 0.8


    def __str__(self):
        output = super(PlaneWaveTest, self).__str__()
        output += "  Eigen Method = %s\n" % self.rundata.multilayer_data.eigen_method
        return output


# Defintion of tests defined here
tests = []
tests.append(PlaneWaveTest())
tests.append(PlaneWaveTest(numpy.pi / 4.0, numpy.pi / 4.0))
tests.append(PlaneWaveTest(numpy.pi / 4.0, 0.0))
tests.append(PlaneWaveTest(0.0, numpy.pi / 4.0))
tests.append(PlaneWaveTest(numpy.pi / 8.0, 0.0))
tests.append(PlaneWaveTest(numpy.pi / 4.0, numpy.pi / 8.0))
for method in [1,2,3,4]:
    tests.append(BubbleTest(eigen_method=method))


if __name__ == "__main__":

    if len(sys.argv) > 1:
        if sys.argv[1].lower() == 'all':
            tests_to_run = tests
        else:
            tests_to_run = []
            for test in sys.argv[1:]:
                tests_to_run.append(tests[int(test)])

        controller = batch.BatchController(tests_to_run)
        print(controller)
        controller.run()

    else:
        controller = batch.BatchController(tests)
        print(controller)





#!/usr/bin/env python
# encoding: utf-8

r"""
Shallow water flow
==================

Solve the one-dimensional shallow water equations including bathymetry:

.. math::
    h_t + (hu)_x & = 0 \\
    (hu)_t + (hu^2 + \frac{1}{2}gh^2)_x & = -g h b_x.

Here h is the depth, u is the velocity, g is the gravitational constant, and b
the bathymetry.  
"""

import numpy
from clawpack import riemann

def setup(kernel_language='Fortran', solver_type='classic', use_petsc=False,
          outdir='./_output'):

    if use_petsc:
        import clawpack.petclaw as pyclaw
    else:
        from clawpack import pyclaw

    if kernel_language == 'Fortran':
        solver = pyclaw.ClawSolver1D(riemann.shallow_bathymetry_fwave_1D)
    elif kernel_language == 'Python':
        solver = pyclaw.ClawSolver1D(riemann.shallow_1D_py.shallow_fwave_1d)
        solver.kernel_language = 'Python'
    solver.limiters = pyclaw.limiters.tvd.vanleer
    solver.fwave = True
    solver.num_waves = 2
    solver.num_eqn = 2
    solver.bc_lower[0] = pyclaw.BC.extrap
    solver.bc_upper[0] = pyclaw.BC.extrap
    solver.aux_bc_lower[0] = pyclaw.BC.extrap
    solver.aux_bc_upper[0] = pyclaw.BC.extrap

    xlower = -1.0
    xupper = 1.0
    x = pyclaw.Dimension( xlower, xupper, 500, name='x')
    domain = pyclaw.Domain(x)
    state = pyclaw.State(domain, 2, 1)

    # Gravitational constant
    state.problem_data['grav'] = 9.8
    state.problem_data['dry_tolerance'] = 1e-3
    state.problem_data['sea_level'] = 0.0

    xc = state.grid.x.centers
    state.aux[0, :] = 0.8 * numpy.exp(-xc**2 / 0.2**2) - 1.0
    state.q[0, :] = 0.1 * numpy.exp(-(xc + 0.4)**2 / 0.2**2) - state.aux[0, :]
    state.q[1, :] = 0.0

    claw = pyclaw.Controller()
    claw.keep_copy = True
    claw.tfinal = 1.0
    claw.solution = pyclaw.Solution(state, domain)
    claw.solver = solver
    claw.setplot = setplot
    claw.write_aux_init = True

    if outdir is not None:
        claw.outdir = outdir
    else:
        claw.output_format = None


    return claw


#--------------------------
def setplot(plotdata):
#--------------------------
    """ 
    Specify what is to be plotted at each frame.
    Input:  plotdata, an instance of visclaw.data.ClawPlotData.
    Output: a modified version of plotdata.
    """ 
    plotdata.clearfigures()  # clear any old figures,axes,items data

    # Plot variables
    def bathy(current_data):
        return current_data.aux[0, :]

    def eta(current_data):
        return current_data.q[0, :] + bathy(current_data)

    def velocity(current_data):
        return current_data.q[1, :] / current_data.q[0, :]

    rgb_converter = lambda triple: [float(rgb) / 255.0 for rgb in triple]

    # Figure for depth
    plotfigure = plotdata.new_plotfigure(name='Depth', figno=0)

    # Axes for water depth
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.xlimits = [-1.0, 1.0]
    plotaxes.ylimits = [-1.1, 0.2]
    plotaxes.title = 'Water Depth'
    plotaxes.axescmd = 'subplot(211)'

    plotitem = plotaxes.new_plotitem(plot_type='1d_fill_between')
    plotitem.plot_var = eta
    plotitem.plot_var2 = bathy
    plotitem.color = rgb_converter((67,183,219))

    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = bathy
    plotitem.color = 'k'

    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = eta
    plotitem.color = 'k'

    # Axes for velocity
    plotaxes = plotfigure.new_plotaxes()
    plotaxes.axescmd = 'subplot(212)'
    plotaxes.xlimits = [-1.0, 1.0]
    plotaxes.ylimits = [-0.5, 0.5]
    plotaxes.title = 'Velocity'

    plotitem = plotaxes.new_plotitem(plot_type='1d_plot')
    plotitem.plot_var = velocity
    plotitem.color = 'b'
    plotitem.kwargs = {'linewidth':3}
    
    return plotdata


if __name__=="__main__":
    from clawpack.pyclaw.util import run_app_from_main
    output = run_app_from_main(setup,setplot)



""" 
Module to set up run time parameters for Clawpack -- AMRClaw code.

The values set in the function setrun are then written out to data files
that will be read in by the Fortran code.
    
""" 

from __future__ import print_function
import os, sys
import numpy as np

from clawpack.geoclaw.data import Rearth  # radius of earth
from mapper import latlong

from clawpack.geoclaw import fgmax_tools

#------------------------------
def setrun(claw_pkg='geoclaw'):
#------------------------------
    
    """ 
    Define the parameters used for running Clawpack.

    INPUT:
        claw_pkg expected to be "geoclaw" for this setrun.

    OUTPUT:
        rundata - object of class ClawRunData 
    
    """ 
    
    from clawpack.clawutil import data 
    
    
    assert claw_pkg.lower() == 'geoclaw',  "Expected claw_pkg = 'geoclaw'"

    num_dim = 2
    rundata = data.ClawRunData(claw_pkg, num_dim)


    #------------------------------------------------------------------
    # Problem-specific parameters to be written to setprob.data:
    #------------------------------------------------------------------
    theta_island = 220.
    probdata = rundata.new_UserData(name='probdata',fname='setprob.data')
    probdata.add_param('theta_island', theta_island,  'angle to island')

    
    #------------------------------------------------------------------
    # Standard Clawpack parameters to be written to claw.data:
    #------------------------------------------------------------------

    clawdata = rundata.clawdata  # initialized when rundata instantiated


    # Set single grid parameters first.
    # See below for AMR parameters.


    # ---------------
    # Spatial domain:
    # ---------------

    # Number of space dimensions:
    clawdata.num_dim = num_dim
    
    # Lower and upper edge of computational domain:
    clawdata.lower[0] = -20.0          # xlower
    clawdata.upper[0] = 20.0          # xupper
    clawdata.lower[1] = 20.0          # ylower
    clawdata.upper[1] = 60.0          # yupper
    
    # Number of grid cells:
    clawdata.num_cells[0] = 40      # mx
    clawdata.num_cells[1] = 40      # my
    

    # ---------------
    # Size of system:
    # ---------------

    # Number of equations in the system:
    clawdata.num_eqn = 3

    # Number of auxiliary variables in the aux array (initialized in setaux)
    clawdata.num_aux = 3
    
    # Index of aux array corresponding to capacity function, if there is one:
    clawdata.capa_index = 2
    
    
    # -------------
    # Initial time:
    # -------------

    clawdata.t0 = 0.0
    

    # Restart from checkpoint file of a previous run?
    # Note: If restarting, you must also change the Makefile to set:
    #    RESTART = True
    # If restarting, t0 above should be from original run, and the
    # restart_file 'fort.chkNNNNN' specified below should be in 
    # the OUTDIR indicated in Makefile.

    clawdata.restart = False               # True to restart from prior results
    clawdata.restart_file = 'fort.chk00006'  # File to use for restart data
    
    
    # -------------
    # Output times:
    #--------------

    # Specify at what times the results should be written to fort.q files.
    # Note that the time integration stops after the final output time.
 
    clawdata.output_style = 1
 
    if clawdata.output_style==1:
        # Output ntimes frames at equally spaced times up to tfinal:
        # Can specify num_output_times = 0 for no output
        clawdata.num_output_times = 7
        clawdata.tfinal = 14000.
        clawdata.output_t0 = True  # output at initial (or restart) time?
        
    elif clawdata.output_style == 2:
        # Specify a list or numpy array of output times:
        # Include t0 if you want output at the initial time.
        clawdata.output_times =  [0., 0.1]
 
    elif clawdata.output_style == 3:
        # Output every step_interval timesteps over total_steps timesteps:
        clawdata.output_step_interval = 2
        clawdata.total_steps = 4
        clawdata.output_t0 = True  # output at initial (or restart) time?
        

    clawdata.output_format = 'binary'      # 'ascii', 'binary'

    clawdata.output_q_components = 'all'   # could be list such as [True,True]
    clawdata.output_aux_components = 'none'  # could be list
    clawdata.output_aux_onlyonce = True    # output aux arrays only at t0
    

    # ---------------------------------------------------
    # Verbosity of messages to screen during integration:  
    # ---------------------------------------------------

    # The current t, dt, and cfl will be printed every time step
    # at AMR levels <= verbosity.  Set verbosity = 0 for no printing.
    #   (E.g. verbosity == 2 means print only on levels 1 and 2.)
    clawdata.verbosity = 1
    
    

    # --------------
    # Time stepping:
    # --------------

    # if dt_variable==True:  variable time steps used based on cfl_desired,
    # if dt_variable==Falseixed time steps dt = dt_initial always used.
    clawdata.dt_variable = True
    
    # Initial time step for variable dt.  
    # (If dt_variable==0 then dt=dt_initial for all steps)
    clawdata.dt_initial = 16.0
    
    # Max time step to be allowed if variable dt used:
    clawdata.dt_max = 1e+99
    
    # Desired Courant number if variable dt used 
    clawdata.cfl_desired = 0.75
    # max Courant number to allow without retaking step with a smaller dt:
    clawdata.cfl_max = 1.0
    
    # Maximum number of time steps to allow between output times:
    clawdata.steps_max = 5000


    # ------------------
    # Method to be used:
    # ------------------

    # Order of accuracy:  1 => Godunov,  2 => Lax-Wendroff plus limiters
    clawdata.order = 2
    
    # Use dimensional splitting? (not yet available for AMR)
    clawdata.dimensional_split = 'unsplit'
    
    # For unsplit method, transverse_waves can be 
    #  0 or 'none'      ==> donor cell (only normal solver used)
    #  1 or 'increment' ==> corner transport of waves
    #  2 or 'all'       ==> corner transport of 2nd order corrections too
    clawdata.transverse_waves = 2
    
    
    # Number of waves in the Riemann solution:
    clawdata.num_waves = 3
    
    # List of limiters to use for each wave family:  
    # Required:  len(limiter) == num_waves
    # Some options:
    #   0 or 'none'     ==> no limiter (Lax-Wendroff)
    #   1 or 'minmod'   ==> minmod
    #   2 or 'superbee' ==> superbee
    #   3 or 'vanleer'  ==> van Leer
    #   4 or 'mc'       ==> MC limiter
    clawdata.limiter = ['vanleer', 'vanleer', 'vanleer']
    
    clawdata.use_fwaves = True    # True ==> use f-wave version of algorithms
    
    # Source terms splitting:
    #   src_split == 0 or 'none'    ==> no source term (src routine never called)
    #   src_split == 1 or 'godunov' ==> Godunov (1st order) splitting used, 
    #   src_split == 2 or 'strang'  ==> Strang (2nd order) splitting used,  not recommended.
    clawdata.source_split = 1
    
    
    # --------------------
    # Boundary conditions:
    # --------------------

    # Number of ghost cells (usually 2)
    clawdata.num_ghost = 2
    
    # Choice of BCs at xlower and xupper:
    #   0 or 'user'     => user specified (must modify bcNamr.f to use this option)
    #   1 or 'extrap'   => extrapolation (non-reflecting outflow)
    #   2 or 'periodic' => periodic (must specify this at both boundaries)
    #   3 or 'wall'     => solid wall for systems where q(2) is normal velocity
    
    clawdata.bc_lower[0] = 'extrap'   # at xlower
    clawdata.bc_upper[0] = 'extrap'   # at xupper

    clawdata.bc_lower[1] = 'extrap'   # at ylower
    clawdata.bc_upper[1] = 'extrap'   # at yupper
                  
       
    # ---------------
    # Gauges:
    # ---------------
    gauges  = rundata.gaugedata.gauges 
    # for gauges append lines of the form  [gaugeno, x, y, t1, t2]

    gaugeno = 0
    for d in [1570e3, 1590e3, 1610e3, 1630e3]:
        gaugeno = gaugeno+1
        x,y = latlong(d, theta_island, 40., Rearth)
        gauges.append([gaugeno, x, y, 0., 1e10])

                  
    # --------------
    # Checkpointing:
    # --------------

    # Specify when checkpoint files should be created that can be
    # used to restart a computation.

    clawdata.checkpt_style = 1

    if clawdata.checkpt_style == 0:
      # Do not checkpoint at all
      pass

    elif clawdata.checkpt_style == 1:
      # Checkpoint only at tfinal.
      pass

    elif clawdata.checkpt_style == 2:
      # Specify a list of checkpoint times.  
      clawdata.checkpt_times = [0.1,0.15]

    elif clawdata.checkpt_style == 3:
      # Checkpoint every checkpt_interval timesteps (on Level 1)
      # and at the final time.
      clawdata.checkpt_interval = 5

    

    # ---------------
    # AMR parameters:   (written to amr.data)
    # ---------------
    amrdata = rundata.amrdata

    # max number of refinement levels:
    amrdata.amr_levels_max = 5

    # List of refinement ratios at each level (length at least amr_level_max-1)
    amrdata.refinement_ratios_x = [4, 3, 5, 4]
    amrdata.refinement_ratios_y = [4, 3, 5, 4]
    amrdata.refinement_ratios_t = [1, 1, 1, 1]


    # Specify type of each aux variable in amrdata.auxtype.
    # This must be a list of length num_aux, each element of which is one of:
    #   'center',  'capacity', 'xleft', or 'yleft'  (see documentation).
    amrdata.aux_type = ['center', 'capacity', 'yleft']


    # Flag for refinement based on Richardson error estimater:
    amrdata.flag_richardson = False    # use Richardson?
    amrdata.flag_richardson_tol = 1.0  # Richardson tolerance
    
    # Flag for refinement using routine flag2refine:
    amrdata.flag2refine = True      # use this?
    amrdata.flag2refine_tol = 0.5  # tolerance used in this routine
    # Note: in geoclaw the refinement tolerance is set as wave_tolerance below 
    # and flag2refine_tol is unused!

    # steps to take on each level L between regriddings of level L+1:
    amrdata.regrid_interval = 3       

    # width of buffer zone around flagged points:
    # (typically the same as regrid_interval so waves don't escape):
    amrdata.regrid_buffer_width  = 2

    # clustering alg. cutoff for (# flagged pts) / (total # of cells refined)
    # (closer to 1.0 => more small grids may be needed to cover flagged cells)
    amrdata.clustering_cutoff = 0.7

    # print info about each regridding up to this level:
    amrdata.verbosity_regrid = 0      


    # ---------------
    # Regions:
    # ---------------
    regions = rundata.regiondata.regions
    # to specify regions of refinement append lines of the form
    #  [minlevel,maxlevel,t1,t2,x1,x2,y1,y2]

    # default on full domain (no longer implicit in topo file, as of v5.8.0):
    regions.append([1, 1,    0., 1e9, -20., 20., 20., 60.])
    # initially around hump (no longer implicit in qinit file, as of v5.8.0):
    regions.append([1, 2,    0., 1., -20., 20., 20., 60.])

    regions.append([1, 3,    0., 5000., -5., 20., 35., 55.])
    regions.append([1, 2, 5000., 6900., -5., 20., 35., 55.])
    regions.append([1, 3, 5000., 6900., 10., 20., 45., 55.])
    regions.append([1, 2, 6900., 9000., 10., 20., 47., 52.])

    # Force refinement near the island as the wave approaches:

    (xisland,yisland) = latlong(1600.e3, theta_island, 40., Rearth)
    x1 = xisland - 1.
    x2 = xisland + 1.
    y1 = yisland - 1.
    y2 = yisland + 1.
    regions.append([4, 4, 7000., 1.e10,  x1,x2,y1,y2])

    x1 = xisland - 0.2
    x2 = xisland + 0.2
    y1 = yisland - 0.2
    y2 = yisland + 0.2
    regions.append([4, 5, 8000., 1.e10,  x1,x2,y1,y2])

    # -----------------------------------------------
    # GeoClaw specific parameters:

    try:
        geo_data = rundata.geo_data
    except:
        print("*** Error, this rundata has no geo_data attribute")
        raise AttributeError("Missing geo_data attribute")

    # == Physics ==
    geo_data.gravity = 9.81
    geo_data.coordinate_system =  2
    geo_data.earth_radius = 6367500.0

    # == Forcing Options
    geo_data.coriolis_forcing = False

    # == Algorithm and Initial Conditions ==
    geo_data.sea_level = 0.0
    geo_data.dry_tolerance = 0.001
    geo_data.friction_forcing = True
    geo_data.manning_coefficient = 0.025
    geo_data.friction_depth = 1000000.0

    # Refinement settings
    refinement_data = rundata.refinement_data
    refinement_data.variable_dt_refinement_ratios = True
    refinement_data.wave_tolerance = 0.01

    # == settopo.data values ==
    topofiles = rundata.topo_data.topofiles
    # for topography, append lines of the form
    #    [topotype, fname]
    topofiles.append([3, 'ocean.tt3'])
    topofiles.append([3, 'island.tt3'])

    # == setdtopo.data values ==
    rundata.dtopo_data.dtopofiles = []
    dtopofiles = rundata.dtopo_data.dtopofiles
    # for moving topography, append lines of the form :  
    #   [topotype, fname]


    # == setqinit.data values ==
    rundata.qinit_data.qinit_type =  4
    qinitfiles = rundata.qinit_data.qinitfiles 
    # for qinit perturbations, append lines of the form: (<= 1 allowed for now!)
    #   [fname]
    rundata.qinit_data.qinitfiles = [['hump.xyz']]


    # == fgmax_grids.data values ==
    # NEW STYLE STARTING IN v5.7.0

    # set num_fgmax_val = 1 to save only max depth,
    #                     2 to also save max speed,
    #                     5 to also save max hs,hss,hmin
    rundata.fgmax_data.num_fgmax_val = 2  # Save depth and speed

    fgmax_grids = rundata.fgmax_data.fgmax_grids  # empty list to start

    # Now append to this list objects of class fgmax_tools.FGmaxGrid
    # specifying any fgmax grids.
    
    # Here several different ones are specified to illustrate:
    
    tstart_max = 8000.  # when to start monitoring fgmax grids

    # Points on a uniform 2d grid:
    fg = fgmax_tools.FGmaxGrid()
    fg.point_style = 2  # uniform rectangular x-y grid  
    fg.x1 = 14.25
    fg.x2 = 14.65
    fg.y1 = 50.10
    fg.y2 = 50.35
    fg.dx = 15/ 3600.  # desired resolution of fgmax grid
    fg.min_level_check = amrdata.amr_levels_max # which levels to monitor max on
    fg.tstart_max = tstart_max  # just before wave arrives
    fg.tend_max = 1.e10    # when to stop monitoring max values
    fg.dt_check = 20.      # how often to update max values
    fg.interp_method = 0   # 0 ==> pw const in cells, recommended
    fgmax_grids.append(fg)  # written to fgmax_grids.data

    # Points on a 1d transect from (x1,y1) to (x2,y2):
    fg = fgmax_tools.FGmaxGrid()
    fg.point_style = 1  # equally spaced points on a transect
    fg.x1 = 14.25
    fg.x2 = 14.65
    fg.y1 = 50.10
    fg.y2 = 50.35
    fg.npts = 50
    fg.min_level_check = amrdata.amr_levels_max # which levels to monitor max on
    fg.tstart_max = tstart_max  # just before wave arrives
    fg.tend_max = 1.e10    # when to stop monitoring max values
    fg.dt_check = 20.      # how often to update max values
    fg.interp_method = 0   # 0 ==> pw const in cells, recommended
    fgmax_grids.append(fg)  # written to fgmax_grids.data

    # fgmax grid point_style==4 means grid specified as topo_type==3 file:
    fg = fgmax_tools.FGmaxGrid()
    fg.point_style = 4
    fg.min_level_check = amrdata.amr_levels_max # which levels to monitor max on
    fg.tstart_max = tstart_max  # just before wave arrives
    fg.tend_max = 1.e10    # when to stop monitoring max values
    fg.dt_check = 20.      # how often to update max values
    fg.interp_method = 0   # 0 ==> pw const in cells, recommended
    fg.xy_fname = 'fgmax_pts_island.data'  # file of 0/1 values in tt3 format
    fgmax_grids.append(fg)  # written to fgmax_grids.data

    # fgmax grid point_style==0 means list of points:
    fg = fgmax_tools.FGmaxGrid()
    fg.point_style = 0
    fg.min_level_check = amrdata.amr_levels_max # which levels to monitor max on
    fg.tstart_max = tstart_max  # just before wave arrives
    fg.tend_max = 1.e10    # when to stop monitoring max values
    fg.dt_check = 20.      # how often to update max values
    fg.interp_method = 0   # 0 ==> pw const in cells, recommended
    # can set list of points here:
    fg.npts = 2
    fg.X = np.array([14.4, 14.5])
    fg.Y = np.array([50.13, 50.13])
    fgmax_grids.append(fg)  # written to fgmax_grids.data

    # fgmax grid point_style==0 means list of points:
    fg = fgmax_tools.FGmaxGrid()
    fg.point_style = 0
    fg.min_level_check = amrdata.amr_levels_max # which levels to monitor max on
    fg.tstart_max = tstart_max  # just before wave arrives
    fg.tend_max = 1.e10    # when to stop monitoring max values
    fg.dt_check = 20.      # how often to update max values
    fg.interp_method = 0   # 0 ==> pw const in cells, recommended
    # can specify that list of points is in a different file:
    fg.npts = 0
    fg.xy_fname = 'fgmax_ps0.txt'
    fgmax_grids.append(fg)  # written to fgmax_grids.data
    
    #  ----- For developers ----- 
    # Toggle debugging print statements:
    amrdata.dprint = False      # print domain flags
    amrdata.eprint = False      # print err est flags
    amrdata.edebug = False      # even more err est flags
    amrdata.gprint = False      # grid bisection/clustering
    amrdata.nprint = False      # proper nesting output
    amrdata.pprint = False      # proj. of tagged points
    amrdata.rprint = False      # print regridding summary
    amrdata.sprint = False      # space/memory output
    amrdata.tprint = False      # time step reporting each level
    amrdata.uprint = False      # update/upbnd reporting
    
    return rundata

    # end of function setrun
    # ----------------------


if __name__ == '__main__':
    # Set up run-time parameters and write all data files.
    import sys
    rundata = setrun(*sys.argv[1:])
    rundata.write()
    
    # remake topo in case island location theta_island was changed above:
    #import maketopo
    #maketopo.maketopo()
    #maketopo.makeqinit()
    
